﻿Namespace My
  ' The following events are available for MyApplication:
  ' 
  ' Startup: Raised when the application starts, before the startup form is created.
  ' Shutdown: Raised after all application forms are closed.  This event is not raised if the application terminates abnormally.
  ' UnhandledException: Raised if the application encounters an unhandled exception.
  ' StartupNextInstance: Raised when launching a single-instance application and the application is already active. 
  ' NetworkAvailabilityChanged: Raised when the network connection is connected or disconnected.
  Partial Friend Class MyApplication
    Private Sub MyApplication_NetworkAvailabilityChanged(sender As Object, e As Microsoft.VisualBasic.Devices.NetworkAvailableEventArgs) Handles Me.NetworkAvailabilityChanged
      If e.IsNetworkAvailable Then
        For Each tabItem As TabPage In frmMain.tbsProfiles.TabPages
          If CType(tabItem.Tag, BNUI).Config.AutoConnect Then CType(tabItem.Tag, BNUI).BNCS_Connect()
        Next
      Else
        frmMain.mnuProfileAllDisconnect.PerformClick()
      End If
    End Sub
    Private Sub MyApplication_Shutdown(sender As Object, e As System.EventArgs) Handles Me.Shutdown
      If My.Computer.FileSystem.FileExists(AppTemp & "\Update.zip") And My.Computer.FileSystem.FileExists(System.Windows.Forms.Application.StartupPath & "\EUp.exe") Then Process.Start(System.Windows.Forms.Application.StartupPath & "\EUp.exe", "/tmp=""" & AppTemp & """ /app=""" & System.Windows.Forms.Application.StartupPath & """")
      My.Settings.Save()
    End Sub
    Private Sub MyApplication_Startup(sender As Object, e As Microsoft.VisualBasic.ApplicationServices.StartupEventArgs) Handles Me.Startup
      If My.Computer.FileSystem.FileExists(AppData & "\Settings.bak") Then
        If My.Computer.FileSystem.FileExists(AppData & "\Settings.xml") Then
          My.Computer.FileSystem.DeleteFile(AppData & "\Settings.xml")
          My.Computer.FileSystem.RenameFile(AppData & "\Settings.bak", "Settings.xml")
        Else
          My.Computer.FileSystem.RenameFile(AppData & "\Settings.bak", "Settings.xml")
        End If
      End If
      If My.Settings.MustUpgrade Then
        My.Settings.Upgrade()
        My.Settings.MustUpgrade = False
        My.Settings.Save()
      End If
      If My.Computer.FileSystem.DirectoryExists(AppTemp & "\old") Then My.Computer.FileSystem.DeleteDirectory(AppTemp & "\old", FileIO.DeleteDirectoryOption.DeleteAllContents)
      For Each sEmptyDir As String In (From sDir As String In My.Computer.FileSystem.GetDirectories(AppData) Where IO.Path.GetFileName(sDir).StartsWith("0."))
        My.Computer.FileSystem.DeleteDirectory(sEmptyDir, FileIO.DeleteDirectoryOption.DeleteAllContents)
      Next
    End Sub
    Private Sub MyApplication_UnhandledException(sender As Object, e As Microsoft.VisualBasic.ApplicationServices.UnhandledExceptionEventArgs) Handles Me.UnhandledException
      If e.Exception.Message.Contains("Could not load file or assembly") Then
        MsgBox("A critical file is missing. Please ensure " & Application.Info.ProductName & " has been fully extracted or installed." & vbNewLine & e.Exception.Message, MsgBoxStyle.Critical, "Could not load File or Assembly.")
        e.ExitApplication = True
      Else
        Dim frmError As New Form With
          {
            .FormBorderStyle = Windows.Forms.FormBorderStyle.SizableToolWindow,
            .Text = "Error in " & Application.Info.ProductName,
            .ShowIcon = False,
            .MinimizeBox = False,
            .MaximizeBox = False,
            .Size = New Size(400, 250),
            .BackColor = Color.Black,
            .ForeColor = Color.White,
            .StartPosition = FormStartPosition.CenterParent,
            .MinimumSize = New Size(340, 200),
            .TopMost = True
          }
        Dim pnlError As New TableLayoutPanel With
          {
            .RowCount = 3,
            .ColumnCount = 4,
            .Dock = DockStyle.Fill
          }
        pnlError.ColumnStyles.Add(New ColumnStyle(SizeType.Absolute, 48))
        pnlError.ColumnStyles.Add(New ColumnStyle(SizeType.Percent, 100))
        pnlError.ColumnStyles.Add(New ColumnStyle(SizeType.AutoSize))
        pnlError.ColumnStyles.Add(New ColumnStyle(SizeType.AutoSize))
        pnlError.RowStyles.Add(New RowStyle(SizeType.Absolute, 48))
        pnlError.RowStyles.Add(New RowStyle(SizeType.Percent, 100))
        pnlError.RowStyles.Add(New RowStyle(SizeType.AutoSize))
        Dim sErrorText As String = Application.Info.ProductName & " has Encountered an Error"
        If e.Exception.TargetSite IsNot Nothing Then sErrorText &= " in " & e.Exception.TargetSite.Name
        Dim lblError As New Label With
          {
            .Font = New Font(frmError.Font.FontFamily, 12, FontStyle.Bold),
            .Text = sErrorText,
            .Dock = DockStyle.Fill,
            .AutoEllipsis = True,
            .TextAlign = ContentAlignment.MiddleLeft
          }
        Dim pctError As New PictureBox With
          {
            .Image = SystemIcons.Error.ToBitmap,
            .SizeMode = PictureBoxSizeMode.AutoSize,
            .BackColor = Color.Transparent,
            .Anchor = AnchorStyles.None
          }
        Dim txtError As New TextBox With
          {
            .ReadOnly = True,
            .BorderStyle = TextBoxBorder,
            .BackColor = Color.Black,
            .Text = "Error: " & e.Exception.Message,
            .Dock = DockStyle.Fill,
            .Multiline = True,
            .ScrollBars = ScrollBars.Vertical,
            .ForeColor = Color.White
          }
        txtError.Text = "Error: " & e.Exception.Message
        If Not String.IsNullOrEmpty(e.Exception.StackTrace) Then
          txtError.Text &= vbNewLine & e.Exception.StackTrace.Substring(0, e.Exception.StackTrace.IndexOf(vbCr))
        Else
          If Not String.IsNullOrEmpty(e.Exception.Source) Then
            txtError.Text &= vbNewLine & " @ " & e.Exception.Source
            If e.Exception.TargetSite IsNot Nothing Then txtError.Text &= "." & e.Exception.TargetSite.Name
          Else
            If e.Exception.TargetSite IsNot Nothing Then txtError.Text &= vbNewLine & " @ " & e.Exception.TargetSite.Name
          End If
        End If
        Dim cmdReport As New Button With
          {
            .Text = "Report Error",
            .AutoSize = True,
            .Padding = New Padding(4),
            .Anchor = AnchorStyles.Right,
            .FlatStyle = FlatStyle.Flat
          }
        cmdReport.FlatAppearance.BorderColor = Color.White
        cmdReport.FlatAppearance.MouseDownBackColor = Color.Black
        cmdReport.FlatAppearance.MouseOverBackColor = System.Drawing.Color.FromArgb(CType(CType(64, Byte), Integer), CType(CType(64, Byte), Integer), CType(CType(64, Byte), Integer))
        AddHandler cmdReport.Click, Sub()
                                      frmError.DialogResult = Windows.Forms.DialogResult.OK
                                      frmError.Close()
                                    End Sub
        Dim cmdIgnore As New Button With
          {
            .Text = "Ignore and Continue",
            .AutoSize = True,
            .Padding = New Padding(4),
            .Anchor = AnchorStyles.Right,
            .FlatStyle = FlatStyle.Flat
          }
        cmdIgnore.FlatAppearance.BorderColor = Color.White
        cmdIgnore.FlatAppearance.MouseDownBackColor = Color.Black
        cmdIgnore.FlatAppearance.MouseOverBackColor = System.Drawing.Color.FromArgb(CType(CType(64, Byte), Integer), CType(CType(64, Byte), Integer), CType(CType(64, Byte), Integer))
        AddHandler cmdIgnore.Click, Sub()
                                      frmError.DialogResult = Windows.Forms.DialogResult.Ignore
                                      frmError.Close()
                                    End Sub
        Dim cmdExit As New Button With
          {
            .Text = "Exit Application",
            .AutoSize = True,
            .Padding = New Padding(4),
            .Anchor = AnchorStyles.Right,
            .FlatStyle = FlatStyle.Flat
          }
        cmdExit.FlatAppearance.BorderColor = Color.White
        cmdExit.FlatAppearance.MouseDownBackColor = Color.Black
        cmdExit.FlatAppearance.MouseOverBackColor = System.Drawing.Color.FromArgb(CType(CType(64, Byte), Integer), CType(CType(64, Byte), Integer), CType(CType(64, Byte), Integer))
        AddHandler cmdExit.Click, Sub()
                                    frmError.DialogResult = Windows.Forms.DialogResult.Abort
                                    frmError.Close()
                                  End Sub
        frmError.Controls.Add(pnlError)
        pnlError.Controls.Add(pctError, 0, 0)
        pnlError.Controls.Add(lblError, 1, 0)
        pnlError.Controls.Add(txtError, 1, 1)
        pnlError.Controls.Add(cmdReport, 0, 2)
        pnlError.Controls.Add(cmdIgnore, 2, 2)
        pnlError.Controls.Add(cmdExit, 3, 2)
        pnlError.SetColumnSpan(lblError, 3)
        pnlError.SetColumnSpan(txtError, 3)
        pnlError.SetColumnSpan(cmdReport, 2)
        frmError.AcceptButton = cmdReport
        frmError.CancelButton = cmdIgnore
        My.Computer.Audio.PlaySystemSound(Media.SystemSounds.Hand)
        Select Case frmError.ShowDialog
          Case DialogResult.OK
            e.ExitApplication = False
            Dim sRet As String = MantisReporter.ReportIssue(e.Exception)
            If sRet = "OK" Then
              MessageBox.Show("Thank you for reporting the error." & vbNewLine & "For details, click Help to visit the Bug Report page.", My.Application.Info.Title & " Error Report Sent!", MessageBoxButtons.OK, MessageBoxIcon.Information, MessageBoxDefaultButton.Button2, 0, "http://bugs.realityripple.com/set_project.php?project_id=1", HelpNavigator.Index)
            Else
              Dim sErrRep As String = "http://bugs.realityripple.com/set_project.php?project_id=1&make_default=no&ref=bug_report_page.php"
              sErrRep &= "?platform=" & IIf(Environment.Is64BitProcess, "x64", IIf(Environment.Is64BitOperatingSystem, "x86-64", "x86"))
              sErrRep &= "%2526os=" & DoubleEncode(My.Computer.Info.OSFullName.Trim)
              sErrRep &= "%2526os_build=" & DoubleEncode(My.Computer.Info.OSVersion.Trim)
              Dim sSum As String = e.Exception.Message
              If sSum.Length > 80 Then sSum = sSum.Substring(0, 77) & "..."
              sErrRep &= "%2526summary=" & DoubleEncode(sSum)
              Dim sDesc As String = e.Exception.Message
              If Not String.IsNullOrEmpty(e.Exception.StackTrace) Then
                sDesc &= vbNewLine & e.Exception.StackTrace.Substring(0, e.Exception.StackTrace.IndexOf(vbCr))
              Else
                If Not String.IsNullOrEmpty(e.Exception.Source) Then
                  sDesc &= vbNewLine & " @ " & e.Exception.Source
                  If e.Exception.TargetSite IsNot Nothing Then sDesc &= "." & e.Exception.TargetSite.Name
                Else
                  If e.Exception.TargetSite IsNot Nothing Then sDesc &= vbNewLine & " @ " & e.Exception.TargetSite.Name
                End If
              End If
              sDesc &= vbNewLine & "Version " & My.Application.Info.Version.ToString.Trim
              sErrRep &= "%2526description=" & DoubleEncode(sDesc)
              MessageBox.Show(sRet & vbNewLine & "If you would like to report the error manually, please click the Help button below.", My.Application.Info.Title & " Error Report Failed!", MessageBoxButtons.OK, MessageBoxIcon.Error, MessageBoxDefaultButton.Button2, Nothing, sErrRep, HelpNavigator.Index)
            End If
          Case DialogResult.Ignore : e.ExitApplication = False
          Case DialogResult.Abort : e.ExitApplication = True
          Case DialogResult.Cancel : e.ExitApplication = False
        End Select
      End If
    End Sub
    Private Function DoubleEncode(inString As String) As String
      Return PercentEncode(PercentEncode(inString))
    End Function
    Private Function PercentEncode(inString As String) As String
      Dim sRet As String = String.Empty
      For I As Integer = inString.Length - 1 To 0 Step -1
        Dim iChar As Integer = Asc(inString(I))
        Select Case iChar
          Case 48 To 57, 65 To 90, 97 To 122 : sRet = inString(I) & sRet
          Case 32 : sRet = "+" & sRet
          Case Else : sRet = "%" & PadHex(iChar, 2) & sRet
        End Select
      Next
      Return sRet
    End Function
  End Class
End Namespace