﻿Imports System.IO
Public Class XSHA
  Private Const vA As UInt32 = &H67452301UI
  Private Const vB As UInt32 = &HEFCDAB89UI
  Private Const vC As UInt32 = &H98BADCFEUI
  Private Const vD As UInt32 = &H10325476UI
  Private Const vE As UInt32 = &HC3D2E1F0UI
  Private Const v1 As UInt32 = &H5A827999UI
  Private Const v2 As UInt32 = &H6ED9EBA1UI
  Private Const v3 As UInt32 = &H8F1BBCDCUI '&H70E44324UI
  Private Const v4 As UInt32 = &HCA62C1D6UI '&H359D3E2AUI
  Private Shared Function ROL(val As UInteger, shift As Integer) As UInteger
    shift = shift And &H1F
    val = (val >> (&H20 - shift)) Or (val << shift)
    Return val
  End Function
  Private Shared Function BrokenSHA(src() As Byte) As Byte()
    If src.Length > &H400 Then Throw New ArgumentOutOfRangeException("Input must be 1024 bytes or less!")
    Dim a As UInteger = vA
    Dim b As UInteger = vB
    Dim c As UInteger = vC
    Dim d As UInteger = vD
    Dim e As UInteger = vE
    Dim g As UInteger = 0
    Dim bBuffer(&H3FF) As Byte
    Array.Copy(src, 0, bBuffer, 0, src.Length)
    Using mData As New MemoryStream(bBuffer)
      Using br As New BinaryReader(mData), bw As New BinaryWriter(mData)
        For I As Integer = 0 To &H4F
          If I < &H40 Then
            mData.Seek(I * 4, SeekOrigin.Begin) : Dim buf1 As UInt32 = br.ReadUInt32
            mData.Seek(&H4, SeekOrigin.Current) : Dim buf2 As UInt32 = br.ReadUInt32
            mData.Seek(&H14, SeekOrigin.Current) : Dim buf3 As UInt32 = br.ReadUInt32
            mData.Seek(&H10, SeekOrigin.Current) : Dim buf4 As UInt32 = br.ReadUInt32
            mData.Seek(&H8, SeekOrigin.Current) : bw.Write(ROL(1, (buf1 Xor buf3 Xor buf2 Xor buf4) Mod &H20))
            mData.Seek(I * 4, IO.SeekOrigin.Begin)
          End If
          Dim f, k As UInteger
          Select Case I
            Case Is < &H14 : f = ((b And c) Or (Not b And d)) : k = v1
            Case Is < &H28 : f = (d Xor c Xor b) : k = v2
            Case Is < &H3C : f = ((c And b) Or (d And c) Or (d And b)) : k = v3
            Case Else : f = (d Xor c Xor b) : k = v4
          End Select
          g = ROL(a, 5) + f + e + br.ReadUInt32 + k
          e = d : d = c : c = ROL(b, &H1E) : b = a : a = g
        Next
      End Using
    End Using
    Using mData As New IO.MemoryStream(&H14)
      Using bw As New IO.BinaryWriter(mData)
        bw.Write(CUInt(vA + a)) : bw.Write(CUInt(vB + b)) : bw.Write(CUInt(vC + c)) : bw.Write(CUInt(vD + d)) : bw.Write(CUInt(vE + e))
      End Using
      Return mData.GetBuffer
    End Using
  End Function
  Public Shared Function CalculateHash(data() As Byte) As Byte()
    Return BrokenSHA(data)
  End Function
  Public Shared Function HashPassword(sPassword As String) As Byte()
    Return BrokenSHA(System.Text.Encoding.GetEncoding(LATIN_1).GetBytes(sPassword.ToLower))
  End Function
  Public Shared Function DoubleHashPassword(sPassword As String, ClientToken As UInt32, ServerToken As UInt32) As Byte()
    Dim ms = New MemoryStream(28)
    Dim bw As New BinaryWriter(ms)
    Dim firstHash() As Byte = HashPassword(sPassword.ToLower)
    bw.Write(ClientToken)
    bw.Write(ServerToken)
    bw.Write(firstHash)
    Dim toCalc() As Byte = ms.GetBuffer
    Return BrokenSHA(toCalc)
  End Function
  Public Shared Function DoubleHashData(data() As Byte, ClientToken As UInt32, ServerToken As UInt32) As Byte()
    Dim ms = New MemoryStream(28)
    Dim bw As New BinaryWriter(ms)
    Dim firstHash() As Byte = BrokenSHA(data)
    bw.Write(ClientToken)
    bw.Write(ServerToken)
    bw.Write(firstHash)
    Dim toCalc() As Byte = ms.GetBuffer
    Return BrokenSHA(toCalc)
  End Function
End Class