﻿Public Class clsChess
  Public Structure PiecePos
    Dim Row As SByte
    Dim Col As SByte
    Sub New(bRow As SByte, cCol As Char)
      Row = bRow
      Col = Asc(Char.ToUpper(cCol)) - 64
    End Sub
    Sub New(bRow As SByte, bCol As SByte)
      Row = bRow
      Col = bCol
    End Sub
    Sub New(Position As String)
      Me.New(Val(Position(1)), Position(0))
    End Sub
    Sub New(BitPos As UInt64)
      Dim iLoops, pVal As UInt64
      pVal = BitPos
      Do While pVal > 1
        pVal = pVal >> 1
        iLoops += 1
      Loop
      Row = iLoops \ 8 + 1
      Col = 8 - (iLoops Mod 8)
    End Sub
    Function ToBitBoard() As UInt64
      Dim iShift As UInt32 = (Row - 1) * 8
      iShift += (8 - Col)
      Return 1UL << iShift
    End Function
    Public Overrides Function Equals(obj As Object) As Boolean
      Try
        Return Row = obj.row And Col = obj.Col
      Catch ex As Exception
        Return False
      End Try
    End Function
    Function sCol() As Char
      Return Chr(Col + 64)
    End Function
    Public Overrides Function ToString() As String
      Return sCol() & Row.ToString
    End Function
  End Structure
  Public Enum PieceType
    Pawn = 0
    Knight = 1
    Rook = 2
    Bishop = 3
    Queen = 4
    King = 5
    [Nothing] = -1
  End Enum
  <Flags()>
  Enum MoveTypes
    Invalid = &H0
    Move = &H1
    Castle = &H2
    Promote = &H4
    Capture = &H8
    Check = &H10
    Checkmate = &H20
    Draw = &H40
  End Enum
  Structure BoardData
    Dim ubBlack As UInt64
    Dim ubWhite As UInt64
    Dim ubBEnP As UInt64
    Dim ubWEnP As UInt64
    Dim ubPawn As UInt64
    Dim ubRook As UInt64
    Dim ubKnight As UInt64
    Dim ubBishop As UInt64
    Dim ubQueen As UInt64
    Dim ubKing As UInt64
    Dim ubStart As UInt64
    Public Overrides Function Equals(obj As Object) As Boolean
      Try
        If Not obj.ubBlack = ubBlack Then Return False
        If Not obj.ubWhite = ubWhite Then Return False
        If Not obj.ubBEnP = ubBEnP Then Return False
        If Not obj.ubWEnP = ubWEnP Then Return False
        If Not obj.ubPawn = ubPawn Then Return False
        If Not obj.ubRook = ubRook Then Return False
        If Not obj.ubKnight = ubKnight Then Return False
        If Not obj.ubBishop = ubBishop Then Return False
        If Not obj.ubQueen = ubQueen Then Return False
        If Not obj.ubKing = ubKing Then Return False
        If Not obj.ubStart = ubStart Then Return False
        Return True
      Catch ex As Exception
        Return False
      End Try
    End Function
    Public Function IsEmpty() As Boolean
      Return (ubBlack = 0 And ubWhite = 0 And ubBEnP = 0 And ubWEnP = 0 And ubPawn = 0 And ubRook = 0 And ubKnight = 0 And ubBishop = 0 And ubQueen = 0 And ubKing = 0 And ubStart = 0)
    End Function
  End Structure
  Private PieceData As BoardData
  Public cWhiteTiles, cBlackTiles As Color
  Public Sub New()
    PieceData.ubWhite = &HFFFFUL
    PieceData.ubBlack = &HFFFF000000000000UL
    PieceData.ubPawn = &HFF00000000FF00UL
    PieceData.ubRook = &H8100000000000081UL
    PieceData.ubKnight = &H4200000000000042UL
    PieceData.ubBishop = &H2400000000000024UL
    PieceData.ubQueen = &H1000000000000010UL
    PieceData.ubKing = &H800000000000008UL
    PieceData.ubStart = PieceData.ubRook Or PieceData.ubKing
    PieceData.ubBEnP = 0UL
    PieceData.ubWEnP = 0UL
  End Sub
#Region "Movement Functions"
#Region "Fields of Movement"
#Region "Pawn"
  Private Function FieldPawn_Move(fromPoint As PiecePos, isWhite As Boolean, bData As BoardData) As UInt64
    Dim uRet As UInt64
    If isWhite Then
      If fromPoint.Row = 8 Then Return 0
      uRet = fromPoint.ToBitBoard << 8
      If fromPoint.Row = 2 And Not CBool((fromPoint.ToBitBoard << 8) And (bData.ubBlack Or bData.ubWhite)) Then uRet = uRet Or (fromPoint.ToBitBoard << 16)
      uRet = uRet And Not (bData.ubBlack Or bData.ubWhite)
    Else
      If fromPoint.Row = 1 Then Return 0
      uRet = fromPoint.ToBitBoard >> 8
      If fromPoint.Row = 7 And Not CBool((fromPoint.ToBitBoard >> 8) And (bData.ubBlack Or bData.ubWhite)) Then uRet = uRet Or (fromPoint.ToBitBoard >> 16)
      uRet = uRet And Not (bData.ubBlack Or bData.ubWhite)
    End If
    Return uRet
  End Function
  Private Function FieldPawn_Attack(fromPoint As PiecePos, isWhite As Boolean, bData As BoardData, checkForEnemy As Boolean) As UInt64
    Dim uRet As UInt64
    If isWhite Then
      If fromPoint.Row = 8 Then Return 0
      If fromPoint.Col < 8 Then uRet = uRet Or (fromPoint.ToBitBoard << 7)
      If fromPoint.Col > 1 Then uRet = uRet Or (fromPoint.ToBitBoard << 9)
      uRet = uRet And Not bData.ubWhite
      If checkForEnemy Then uRet = uRet And (bData.ubBlack Or bData.ubBEnP)
    Else
      If fromPoint.Row = 8 Then Return 0
      If fromPoint.Col > 1 Then uRet = uRet Or (fromPoint.ToBitBoard >> 7)
      If fromPoint.Col < 8 Then uRet = uRet Or (fromPoint.ToBitBoard >> 9)
      uRet = uRet And Not bData.ubBlack
      If checkForEnemy Then uRet = uRet And (bData.ubWhite Or bData.ubWEnP)
    End If
    Return uRet
  End Function
  Private Function FieldPawn(fromPoint As PiecePos, isWhite As Boolean, bData As BoardData, checkForEnemy As Boolean) As UInt64
    Return FieldPawn_Move(fromPoint, isWhite, bData) Or FieldPawn_Attack(fromPoint, isWhite, bData, checkForEnemy)
  End Function
#End Region
#Region "Rook"
  Private Function FieldRook_Move(fromPoint As PiecePos, isWhite As Boolean, bData As BoardData) As UInt64
    Dim uRet As UInt64
    If fromPoint.Row > 1 Then
      For Y As Integer = fromPoint.Row - 1 To 1 Step -1
        Dim pHere As New PiecePos(Y, fromPoint.Col)
        If CBool(pHere.ToBitBoard And (bData.ubBlack Or bData.ubWhite)) Then Exit For
        uRet = uRet Or pHere.ToBitBoard
      Next
    End If
    If fromPoint.Row < 8 Then
      For Y As Integer = fromPoint.Row + 1 To 8
        Dim pHere As New PiecePos(Y, fromPoint.Col)
        If CBool(pHere.ToBitBoard And (bData.ubBlack Or bData.ubWhite)) Then Exit For
        uRet = uRet Or pHere.ToBitBoard
      Next
    End If
    If fromPoint.Col > 1 Then
      For X As Integer = fromPoint.Col - 1 To 1 Step -1
        Dim pHere As New PiecePos(fromPoint.Row, X)
        If CBool(pHere.ToBitBoard And (bData.ubBlack Or bData.ubWhite)) Then Exit For
        uRet = uRet Or pHere.ToBitBoard
      Next
    End If
    If fromPoint.Col < 8 Then
      For X As Integer = fromPoint.Col + 1 To 8
        Dim pHere As New PiecePos(fromPoint.Row, X)
        If CBool(pHere.ToBitBoard And (bData.ubBlack Or bData.ubWhite)) Then Exit For
        uRet = uRet Or pHere.ToBitBoard
      Next
    End If
    Return uRet And Not fromPoint.ToBitBoard
  End Function
  Private Function FieldRook_Attack(fromPoint As PiecePos, isWhite As Boolean, bData As BoardData, checkForEnemy As Boolean) As UInt64
    Dim uRet As UInt64
    If fromPoint.Row > 1 Then
      For Y As Integer = fromPoint.Row - 1 To 1 Step -1
        Dim pHere As New PiecePos(Y, fromPoint.Col)
        If CBool(pHere.ToBitBoard And (IIf(isWhite, bData.ubWhite, bData.ubBlack))) Then Exit For
        uRet = uRet Or pHere.ToBitBoard
        If CBool(pHere.ToBitBoard And (IIf(isWhite, bData.ubBlack, bData.ubWhite))) Then Exit For
      Next
    End If
    If fromPoint.Row < 8 Then
      For Y As Integer = fromPoint.Row + 1 To 8
        Dim pHere As New PiecePos(Y, fromPoint.Col)
        If CBool(pHere.ToBitBoard And IIf(isWhite, bData.ubWhite, bData.ubBlack)) Then Exit For
        uRet = uRet Or pHere.ToBitBoard
        If CBool(pHere.ToBitBoard And (IIf(isWhite, bData.ubBlack, bData.ubWhite))) Then Exit For
      Next
    End If
    If fromPoint.Col > 1 Then
      For X As Integer = fromPoint.Col - 1 To 1 Step -1
        Dim pHere As New PiecePos(fromPoint.Row, X)
        If CBool(pHere.ToBitBoard And IIf(isWhite, bData.ubWhite, bData.ubBlack)) Then Exit For
        uRet = uRet Or pHere.ToBitBoard
        If CBool(pHere.ToBitBoard And (IIf(isWhite, bData.ubBlack, bData.ubWhite))) Then Exit For
      Next
    End If
    If fromPoint.Col < 8 Then
      For X As Integer = fromPoint.Col + 1 To 8
        Dim pHere As New PiecePos(fromPoint.Row, X)
        If CBool(pHere.ToBitBoard And IIf(isWhite, bData.ubWhite, bData.ubBlack)) Then Exit For
        uRet = uRet Or pHere.ToBitBoard
        If CBool(pHere.ToBitBoard And (IIf(isWhite, bData.ubBlack, bData.ubWhite))) Then Exit For
      Next
    End If
    If checkForEnemy Then
      If isWhite Then
        uRet = uRet And bData.ubBlack
      Else
        uRet = uRet And bData.ubWhite
      End If
    End If
    Return uRet And Not fromPoint.ToBitBoard
  End Function
  Private Function FieldRook(fromPoint As PiecePos, isWhite As Boolean, bData As BoardData, checkForEnemy As Boolean) As UInt64
    Return FieldRook_Move(fromPoint, isWhite, bData) Or FieldRook_Attack(fromPoint, isWhite, bData, checkForEnemy)
  End Function
#End Region
#Region "Knight"
  Private Function FieldKnight_Move(fromPoint As PiecePos, isWhite As Boolean, bData As BoardData) As UInt64
    Dim fPos As UInt64 = 0
    If fromPoint.Col < 8 And fromPoint.Row > 2 Then fPos = fPos Or (fromPoint.ToBitBoard >> (8 + 8 + 1))
    If fromPoint.Col > 1 And fromPoint.Row > 2 Then fPos = fPos Or (fromPoint.ToBitBoard >> (8 + 8 - 1))
    If fromPoint.Col < 7 And fromPoint.Row > 1 Then fPos = fPos Or (fromPoint.ToBitBoard >> (8 + 2))
    If fromPoint.Col > 2 And fromPoint.Row > 1 Then fPos = fPos Or (fromPoint.ToBitBoard >> (8 - 2))
    If fromPoint.Col < 8 And fromPoint.Row < 7 Then fPos = fPos Or (fromPoint.ToBitBoard << (8 + 8 - 1))
    If fromPoint.Col > 1 And fromPoint.Row < 7 Then fPos = fPos Or (fromPoint.ToBitBoard << (8 + 8 + 1))
    If fromPoint.Col < 7 And fromPoint.Row < 8 Then fPos = fPos Or (fromPoint.ToBitBoard << (8 - 2))
    If fromPoint.Col > 2 And fromPoint.Row < 8 Then fPos = fPos Or (fromPoint.ToBitBoard << (8 + 2))
    Return fPos And Not (bData.ubBlack Or bData.ubWhite)
  End Function
  Private Function FieldKnight_Attack(fromPoint As PiecePos, isWhite As Boolean, bData As BoardData, checkForEnemy As Boolean) As UInt64
    Dim fPos As UInt64 = 0
    If fromPoint.Col < 8 And fromPoint.Row > 2 Then fPos = fPos Or (fromPoint.ToBitBoard >> (8 + 8 + 1))
    If fromPoint.Col > 1 And fromPoint.Row > 2 Then fPos = fPos Or (fromPoint.ToBitBoard >> (8 + 8 - 1))
    If fromPoint.Col < 7 And fromPoint.Row > 1 Then fPos = fPos Or (fromPoint.ToBitBoard >> (8 + 2))
    If fromPoint.Col > 2 And fromPoint.Row > 1 Then fPos = fPos Or (fromPoint.ToBitBoard >> (8 - 2))
    If fromPoint.Col < 8 And fromPoint.Row < 7 Then fPos = fPos Or (fromPoint.ToBitBoard << (8 + 8 - 1))
    If fromPoint.Col > 1 And fromPoint.Row < 7 Then fPos = fPos Or (fromPoint.ToBitBoard << (8 + 8 + 1))
    If fromPoint.Col < 7 And fromPoint.Row < 8 Then fPos = fPos Or (fromPoint.ToBitBoard << (8 - 2))
    If fromPoint.Col > 2 And fromPoint.Row < 8 Then fPos = fPos Or (fromPoint.ToBitBoard << (8 + 2))
    fPos = fPos And Not IIf(isWhite, bData.ubWhite, bData.ubBlack)
    If checkForEnemy Then fPos = fPos And IIf(isWhite, bData.ubBlack, bData.ubWhite)
    Return fPos
  End Function
  Private Function FieldKnight(fromPoint As PiecePos, isWhite As Boolean, bData As BoardData, checkForEnemy As Boolean) As UInt64
    Return FieldKnight_Move(fromPoint, isWhite, bData) Or FieldKnight_Attack(fromPoint, isWhite, bData, checkForEnemy)
  End Function
#End Region
#Region "Bishop"
  Private Function FieldBishop_Move(fromPoint As PiecePos, isWhite As Boolean, bData As BoardData) As UInt64
    Dim uRet As UInt64
    If fromPoint.Row > 1 And fromPoint.Col > 1 Then
      Dim iStart As Integer = 2
      If fromPoint.Row > fromPoint.Col Then
        iStart = fromPoint.Col
      Else
        iStart = fromPoint.Row
      End If
      For I As Integer = 1 To iStart - 1
        Dim pHere As New PiecePos(fromPoint.Row - I, fromPoint.Col - I)
        If CBool(pHere.ToBitBoard And (bData.ubBlack Or bData.ubWhite)) Then Exit For
        uRet = uRet Or pHere.ToBitBoard
      Next
    End If
    If fromPoint.Row > 1 And fromPoint.Col < 9 Then
      Dim iStart As Integer = 2
      If fromPoint.Row > (9 - fromPoint.Col) Then
        iStart = 9 - fromPoint.Col
      Else
        iStart = fromPoint.Row
      End If
      For I As Integer = 1 To iStart - 1
        Dim pHere As New PiecePos(fromPoint.Row - I, fromPoint.Col + I)
        If CBool(pHere.ToBitBoard And (bData.ubBlack Or bData.ubWhite)) Then Exit For
        uRet = uRet Or pHere.ToBitBoard
      Next
    End If
    If fromPoint.Row < 9 And fromPoint.Col < 9 Then
      Dim iStart As Integer = 2
      If (9 - fromPoint.Row) > (9 - fromPoint.Col) Then
        iStart = 9 - fromPoint.Col
      Else
        iStart = 9 - fromPoint.Row
      End If
      For I As Integer = 1 To iStart - 1
        Dim pHere As New PiecePos(fromPoint.Row + I, fromPoint.Col + I)
        If CBool(pHere.ToBitBoard And (bData.ubBlack Or bData.ubWhite)) Then Exit For
        uRet = uRet Or pHere.ToBitBoard
      Next
    End If
    If fromPoint.Row < 9 And fromPoint.Col > 1 Then
      Dim iStart As Integer = 2
      If (9 - fromPoint.Row) > fromPoint.Col Then
        iStart = fromPoint.Col
      Else
        iStart = 9 - fromPoint.Row
      End If
      For I As Integer = 1 To iStart - 1
        Dim pHere As New PiecePos(fromPoint.Row + I, fromPoint.Col - I)
        If CBool(pHere.ToBitBoard And (bData.ubBlack Or bData.ubWhite)) Then Exit For
        uRet = uRet Or pHere.ToBitBoard
      Next
    End If
    Return uRet And Not fromPoint.ToBitBoard
  End Function
  Private Function FieldBishop_Attack(fromPoint As PiecePos, isWhite As Boolean, bData As BoardData, checkForEnemy As Boolean) As UInt64
    Dim uRet As UInt64
    If fromPoint.Row > 1 And fromPoint.Col > 1 Then
      Dim iStart As Integer = 2
      If fromPoint.Row > fromPoint.Col Then
        iStart = fromPoint.Col
      Else
        iStart = fromPoint.Row
      End If
      For I As Integer = 1 To iStart - 1
        Dim pHere As New PiecePos(fromPoint.Row - I, fromPoint.Col - I)
        If CBool(pHere.ToBitBoard And (IIf(isWhite, bData.ubWhite, bData.ubBlack))) Then Exit For
        uRet = uRet Or pHere.ToBitBoard
        If CBool(pHere.ToBitBoard And (IIf(isWhite, bData.ubBlack, bData.ubWhite))) Then Exit For
      Next
    End If
    If fromPoint.Row > 1 And fromPoint.Col < 9 Then
      Dim iStart As Integer = 2
      If fromPoint.Row > (9 - fromPoint.Col) Then
        iStart = 9 - fromPoint.Col
      Else
        iStart = fromPoint.Row
      End If
      For I As Integer = 1 To iStart - 1
        Dim pHere As New PiecePos(fromPoint.Row - I, fromPoint.Col + I)
        If CBool(pHere.ToBitBoard And (IIf(isWhite, bData.ubWhite, bData.ubBlack))) Then Exit For
        uRet = uRet Or pHere.ToBitBoard
        If CBool(pHere.ToBitBoard And (IIf(isWhite, bData.ubBlack, bData.ubWhite))) Then Exit For
      Next
    End If
    If fromPoint.Row < 9 And fromPoint.Col < 9 Then
      Dim iStart As Integer = 2
      If (9 - fromPoint.Row) > (9 - fromPoint.Col) Then
        iStart = 9 - fromPoint.Col
      Else
        iStart = 9 - fromPoint.Row
      End If
      For I As Integer = 1 To iStart - 1
        Dim pHere As New PiecePos(fromPoint.Row + I, fromPoint.Col + I)
        If CBool(pHere.ToBitBoard And (IIf(isWhite, bData.ubWhite, bData.ubBlack))) Then Exit For
        uRet = uRet Or pHere.ToBitBoard
        If CBool(pHere.ToBitBoard And (IIf(isWhite, bData.ubBlack, bData.ubWhite))) Then Exit For
      Next
    End If
    If fromPoint.Row < 9 And fromPoint.Col > 1 Then
      Dim iStart As Integer = 2
      If (9 - fromPoint.Row) > fromPoint.Col Then
        iStart = fromPoint.Col
      Else
        iStart = 9 - fromPoint.Row
      End If
      For I As Integer = 1 To iStart - 1
        Dim pHere As New PiecePos(fromPoint.Row + I, fromPoint.Col - I)
        If CBool(pHere.ToBitBoard And (IIf(isWhite, bData.ubWhite, bData.ubBlack))) Then Exit For
        uRet = uRet Or pHere.ToBitBoard
        If CBool(pHere.ToBitBoard And (IIf(isWhite, bData.ubBlack, bData.ubWhite))) Then Exit For
      Next
    End If
    If checkForEnemy Then
      If isWhite Then
        uRet = uRet And bData.ubBlack
      Else
        uRet = uRet And bData.ubWhite
      End If
    End If
    Return uRet And Not fromPoint.ToBitBoard
  End Function
  Private Function FieldBishop(fromPoint As PiecePos, isWhite As Boolean, bData As BoardData, checkForEnemy As Boolean) As UInt64
    Return FieldBishop_Move(fromPoint, isWhite, bData) Or FieldBishop_Attack(fromPoint, isWhite, bData, checkForEnemy)
  End Function
#End Region
#Region "Queen"
  Private Function FieldQueen_Move(fromPoint As PiecePos, isWhite As Boolean, bData As BoardData) As UInt64
    Return (FieldRook_Move(fromPoint, isWhite, bData) Or FieldBishop_Move(fromPoint, isWhite, bData))
  End Function
  Private Function FieldQueen_Attack(fromPoint As PiecePos, isWhite As Boolean, bData As BoardData, checkForEnemy As Boolean) As UInt64
    Return (FieldRook_Attack(fromPoint, isWhite, bData, checkForEnemy) Or FieldBishop_Attack(fromPoint, isWhite, bData, checkForEnemy))
  End Function
  Private Function FieldQueen(fromPoint As PiecePos, isWhite As Boolean, bData As BoardData, checkForEnemy As Boolean) As UInt64
    Return FieldQueen_Move(fromPoint, isWhite, bData) Or FieldQueen_Attack(fromPoint, isWhite, bData, checkForEnemy)
  End Function
#End Region
#Region "King"
  Private Function FieldKing_Move(fromPoint As PiecePos, isWhite As Boolean, bData As BoardData, IgnoreSafety As Boolean) As UInt64
    Dim fPos As UInt64
    For X As Integer = 1 To 8
      For Y As Integer = 1 To 8
        If Math.Abs(fromPoint.Row - Y) < 2 And Math.Abs(fromPoint.Col - X) < 2 Then
          Dim pHere As New PiecePos(Y, X)
          If CBool(pHere.ToBitBoard And (bData.ubBlack Or bData.ubWhite)) Then Continue For
          If Not IgnoreSafety Then
            If CBool(KingThreats(bData, Not isWhite, True) And pHere.ToBitBoard) Then Continue For
            Dim bNewData As BoardData = bData
            RemovePiece(pHere, bNewData)
            ChangePiece(fromPoint, pHere, bNewData)
            If CBool(KingThreats(bNewData, Not isWhite, True) And pHere.ToBitBoard) Then Continue For
          End If
          fPos = fPos Or pHere.ToBitBoard
        Else
          If ((isWhite And Y = 1) Or (Not isWhite And Y = 8)) And ((X = 3) Or (X = 7)) Then 'check for castle
            Dim bMyColor As UInt64 = IIf(isWhite, bData.ubWhite, bData.ubBlack)
            If Not CBool(bData.ubStart And (bData.ubKing And bMyColor)) Then Continue For 'King has moved
            If Not CBool(bData.ubStart And (bData.ubRook And bMyColor)) Then Continue For 'Rooks have moved
            If X = 3 Then 'queenside
              If Not CBool(bData.ubStart And (New PiecePos(Y, 1).ToBitBoard)) Then Continue For
              If CBool((bData.ubWhite Or bData.ubBlack) And (New PiecePos(Y, 2).ToBitBoard)) Then Continue For
              If CBool((bData.ubWhite Or bData.ubBlack) And (New PiecePos(Y, 3).ToBitBoard)) Then Continue For
              If CBool((bData.ubWhite Or bData.ubBlack) And (New PiecePos(Y, 4).ToBitBoard)) Then Continue For
            Else 'kingside
              If Not CBool(bData.ubStart And (New PiecePos(Y, 8).ToBitBoard)) Then Continue For
              If CBool((bData.ubWhite Or bData.ubBlack) And (New PiecePos(Y, 6).ToBitBoard)) Then Continue For
              If CBool((bData.ubWhite Or bData.ubBlack) And (New PiecePos(Y, 7).ToBitBoard)) Then Continue For
            End If
            If Not IgnoreSafety Then If isInCheck(isWhite, bData) Then Continue For
            Dim pHere As New PiecePos(Y, X)
            fPos = fPos Or pHere.ToBitBoard
          End If
        End If
      Next
    Next
    Return fPos And Not fromPoint.ToBitBoard
  End Function
  Private Function FieldKing_Attack(fromPoint As PiecePos, isWhite As Boolean, bData As BoardData, IgnoreSafety As Boolean) As UInt64
    Dim uRet As UInt64
    For X As Integer = 1 To 8
      For Y As Integer = 1 To 8
        If Math.Abs(fromPoint.Row - Y) < 2 And Math.Abs(fromPoint.Col - X) < 2 Then
          Dim pHere As New PiecePos(Y, X)
          If CBool(pHere.ToBitBoard And IIf(isWhite, bData.ubWhite, bData.ubBlack)) Then Continue For
          uRet = uRet Or pHere.ToBitBoard
          If Not IgnoreSafety Then
            If CBool(KingThreats(bData, Not isWhite, True) And pHere.ToBitBoard) Then Continue For
            Dim bNewData As BoardData = bData
            RemovePiece(pHere, bNewData)
            ChangePiece(fromPoint, pHere, bNewData)
            If CBool(KingThreats(bNewData, Not isWhite, True) And pHere.ToBitBoard) Then Continue For
          End If
        End If
      Next
    Next
    If isWhite Then
      uRet = uRet And bData.ubBlack
    Else
      uRet = uRet And bData.ubWhite
    End If
    Return uRet And Not fromPoint.ToBitBoard
  End Function
  Private Function FieldKing(fromPoint As PiecePos, isWhite As Boolean, bData As BoardData, IgnoreSafety As Boolean) As UInt64
    Return FieldKing_Move(fromPoint, isWhite, bData, IgnoreSafety) Or FieldKing_Attack(fromPoint, isWhite, bData, IgnoreSafety)
  End Function
#End Region
#End Region
#Region "Groups"
  Public Function MovePiece(fromPos As PiecePos, toPos As PiecePos, isSelf As Boolean, myColor As Color, Optional ByRef fromData As BoardData = Nothing, Optional PromoteTo As PieceType = PieceType.Nothing, Optional PromoteSize As Byte = 32) As MoveTypes
    Dim bFromNaught As Boolean = False
    If fromData.IsEmpty Then fromData = PieceData : bFromNaught = True
    Dim mRet As MoveTypes
    If GetPieceColor(fromPos, fromData) = Color.Transparent Then Return MoveTypes.Invalid
    If GetPieceType(fromPos, fromData) = PieceType.Nothing Then Return MoveTypes.Invalid
    If leadsToCheck(fromPos, toPos, fromData) Then Return MoveTypes.Invalid
    Dim isWhite As Boolean = GetPieceColor(fromPos, fromData) = Color.White
    If isSelf Then
      If Not myColor = Color.White And isWhite Then Return MoveTypes.Invalid
      If Not myColor = Color.Black And Not isWhite Then Return MoveTypes.Invalid
    Else
      If myColor = Color.White And isWhite Then Return MoveTypes.Invalid
      If myColor = Color.Black And Not isWhite Then Return MoveTypes.Invalid
    End If
    Select Case GetPieceType(fromPos)
      Case PieceType.Pawn
        If CBool(FieldPawn_Move(fromPos, isWhite, fromData) And toPos.ToBitBoard) Then
          If toPos.Row = fromPos.Row + 2 Then
            fromData.ubWEnP = New PiecePos(fromPos.Row + 1, fromPos.Col).ToBitBoard
          ElseIf toPos.Row = fromPos.Row - 2 Then
            fromData.ubBEnP = New PiecePos(fromPos.Row - 1, fromPos.Col).ToBitBoard
          Else
            fromData.ubBEnP = 0 : fromData.ubWEnP = 0
          End If
          ChangePiece(fromPos, toPos, fromData)
          mRet = mRet Or MoveTypes.Move
          If (isWhite And toPos.Row = 8) Or (Not isWhite And toPos.Row = 1) Then
            If PromoteTo = PieceType.Nothing Then
              Using promote As New frmPromote(isWhite, cBlackTiles, cWhiteTiles, PromoteSize)
                Select Case promote.ShowDialog
                  Case DialogResult.Yes : PromoteTo = PieceType.Queen
                  Case DialogResult.Abort : PromoteTo = PieceType.Bishop
                  Case DialogResult.Retry : PromoteTo = PieceType.Rook
                  Case DialogResult.Ignore : PromoteTo = PieceType.Knight
                  Case DialogResult.Cancel
                End Select
              End Using
            End If
            Select Case PromoteTo
              Case PieceType.Queen
                mRet = mRet Or MoveTypes.Promote
                RemovePiece(toPos, fromData)
                InsertPiece(toPos, isWhite, PieceType.Queen, fromData)
              Case PieceType.Bishop
                mRet = mRet Or MoveTypes.Promote
                RemovePiece(toPos, fromData)
                InsertPiece(toPos, isWhite, PieceType.Bishop, fromData)
              Case PieceType.Rook
                mRet = mRet Or MoveTypes.Promote
                RemovePiece(toPos, fromData)
                InsertPiece(toPos, isWhite, PieceType.Rook, fromData)
              Case PieceType.Knight
                mRet = mRet Or MoveTypes.Promote
                RemovePiece(toPos, fromData)
                InsertPiece(toPos, isWhite, PieceType.Knight, fromData)
            End Select
          End If
        ElseIf CBool(FieldPawn_Attack(fromPos, isWhite, fromData, True) And toPos.ToBitBoard) Then
          If CBool(toPos.ToBitBoard And IIf(isWhite, fromData.ubBEnP, fromData.ubWEnP)) Then 'En Passant
            If isWhite Then
              RemovePiece(New PiecePos(toPos.Row - 1, toPos.Col), fromData)
            Else
              RemovePiece(New PiecePos(toPos.Row + 1, toPos.Col), fromData)
            End If
          Else 'Standard attack
            RemovePiece(toPos, fromData)
          End If
          fromData.ubBEnP = 0 : fromData.ubWEnP = 0
          ChangePiece(fromPos, toPos, fromData)
          mRet = mRet Or MoveTypes.Move Or MoveTypes.Capture
          If (isWhite And toPos.Row = 8) Or (Not isWhite And toPos.Row = 1) Then
            If PromoteTo = PieceType.Nothing Then
              Using promote As New frmPromote(isWhite, cBlackTiles, cWhiteTiles, PromoteSize)
                Select Case promote.ShowDialog
                  Case DialogResult.Yes : PromoteTo = PieceType.Queen
                  Case DialogResult.Abort : PromoteTo = PieceType.Bishop
                  Case DialogResult.Retry : PromoteTo = PieceType.Rook
                  Case DialogResult.Ignore : PromoteTo = PieceType.Knight
                  Case DialogResult.Cancel
                End Select
              End Using
            End If
            Select Case PromoteTo
              Case PieceType.Queen
                mRet = mRet Or MoveTypes.Promote
                RemovePiece(toPos, fromData)
                InsertPiece(toPos, isWhite, PieceType.Queen, fromData)
              Case PieceType.Bishop
                mRet = mRet Or MoveTypes.Promote
                RemovePiece(toPos, fromData)
                InsertPiece(toPos, isWhite, PieceType.Bishop, fromData)
              Case PieceType.Rook
                mRet = mRet Or MoveTypes.Promote
                RemovePiece(toPos, fromData)
                InsertPiece(toPos, isWhite, PieceType.Rook, fromData)
              Case PieceType.Knight
                mRet = mRet Or MoveTypes.Promote
                RemovePiece(toPos, fromData)
                InsertPiece(toPos, isWhite, PieceType.Knight, fromData)
            End Select
          End If
        End If
      Case PieceType.Rook
        If CBool(FieldRook_Move(fromPos, isWhite, fromData) And toPos.ToBitBoard) Then
          ChangePiece(fromPos, toPos, fromData)
          mRet = mRet Or MoveTypes.Move
          fromData.ubBEnP = 0 : fromData.ubWEnP = 0
        ElseIf CBool(FieldRook_Attack(fromPos, isWhite, fromData, True) And toPos.ToBitBoard) Then
          RemovePiece(toPos, fromData)
          ChangePiece(fromPos, toPos, fromData)
          mRet = mRet Or MoveTypes.Move Or MoveTypes.Capture
          fromData.ubBEnP = 0 : fromData.ubWEnP = 0
        End If
      Case PieceType.Knight
        If CBool(FieldKnight_Move(fromPos, isWhite, fromData) And toPos.ToBitBoard) Then
          ChangePiece(fromPos, toPos, fromData)
          mRet = mRet Or MoveTypes.Move
          fromData.ubBEnP = 0 : fromData.ubWEnP = 0
        ElseIf CBool(FieldKnight_Attack(fromPos, isWhite, fromData, True) And toPos.ToBitBoard) Then
          RemovePiece(toPos, fromData)
          ChangePiece(fromPos, toPos, fromData)
          mRet = mRet Or MoveTypes.Move Or MoveTypes.Capture
          fromData.ubBEnP = 0 : fromData.ubWEnP = 0
        End If
      Case PieceType.Bishop
        If CBool(FieldBishop_Move(fromPos, isWhite, fromData) And toPos.ToBitBoard) Then
          ChangePiece(fromPos, toPos, fromData)
          mRet = mRet Or MoveTypes.Move
          fromData.ubBEnP = 0 : fromData.ubWEnP = 0
        ElseIf CBool(FieldBishop_Attack(fromPos, isWhite, fromData, True) And toPos.ToBitBoard) Then
          RemovePiece(toPos, fromData)
          ChangePiece(fromPos, toPos, fromData)
          mRet = mRet Or MoveTypes.Move Or MoveTypes.Capture
          fromData.ubBEnP = 0 : fromData.ubWEnP = 0
        End If
      Case PieceType.Queen
        If CBool(FieldQueen_Move(fromPos, isWhite, fromData) And toPos.ToBitBoard) Then
          ChangePiece(fromPos, toPos, fromData)
          mRet = mRet Or MoveTypes.Move
          fromData.ubBEnP = 0 : fromData.ubWEnP = 0
        ElseIf CBool(FieldQueen_Attack(fromPos, isWhite, fromData, True) And toPos.ToBitBoard) Then
          RemovePiece(toPos, fromData)
          ChangePiece(fromPos, toPos, fromData)
          mRet = mRet Or MoveTypes.Move Or MoveTypes.Capture
          fromData.ubBEnP = 0 : fromData.ubWEnP = 0
        End If
      Case PieceType.King
        If CBool(FieldKing_Move(fromPos, isWhite, fromData, False) And toPos.ToBitBoard) Then
          If Math.Abs(fromPos.Col - toPos.Col) > 1 Then
            ChangePiece(fromPos, toPos, fromData)
            If fromPos.Col > toPos.Col Then ' Queen Side
              ChangePiece(New PiecePos(fromPos.Row, 1), New PiecePos(fromPos.Row, 4), fromData)
            Else 'King Side
              ChangePiece(New PiecePos(fromPos.Row, 8), New PiecePos(fromPos.Row, 6), fromData)
            End If
            mRet = mRet Or MoveTypes.Castle
          Else
            ChangePiece(fromPos, toPos, fromData)
          End If
          mRet = mRet Or MoveTypes.Move
          fromData.ubBEnP = 0 : fromData.ubWEnP = 0
        ElseIf CBool(FieldKing_Attack(fromPos, isWhite, fromData, False) And toPos.ToBitBoard) Then
          RemovePiece(toPos, fromData)
          ChangePiece(fromPos, toPos, fromData)
          mRet = mRet Or MoveTypes.Move Or MoveTypes.Capture
          fromData.ubBEnP = 0 : fromData.ubWEnP = 0
        End If
    End Select
    If isInCheck(Not isWhite, fromData, False) Then mRet = mRet Or MoveTypes.Check
    If bFromNaught Then PieceData = fromData
    Dim bAllowedMovesOfOpponent As New List(Of PiecePos)
    For I As Integer = 0 To 63
      Dim pLoc As New PiecePos(1UL << I)
      If CBool(pLoc.ToBitBoard And IIf(isWhite, fromData.ubBlack, fromData.ubWhite)) Then
        bAllowedMovesOfOpponent.AddRange(GetPieceMoves(pLoc, fromData))
      End If
    Next
    If bAllowedMovesOfOpponent.Count = 0 Then mRet = mRet Or MoveTypes.Checkmate
    Return mRet
  End Function
  Private Sub InsertPiece(toPos As PiecePos, isWhite As Boolean, pType As PieceType, Optional ByRef fromData As BoardData = Nothing)
    Dim bFromNaught As Boolean = False
    If fromData.IsEmpty Then fromData = PieceData : bFromNaught = True

    If isWhite Then
      fromData.ubWhite = fromData.ubWhite Or toPos.ToBitBoard
    Else
      fromData.ubBlack = fromData.ubBlack Or toPos.ToBitBoard
    End If
    Select Case pType
      Case PieceType.Pawn : fromData.ubPawn = fromData.ubPawn Or toPos.ToBitBoard
      Case PieceType.Rook : fromData.ubRook = fromData.ubRook Or toPos.ToBitBoard
      Case PieceType.Knight : fromData.ubKnight = fromData.ubKnight Or toPos.ToBitBoard
      Case PieceType.Bishop : fromData.ubBishop = fromData.ubBishop Or toPos.ToBitBoard
      Case PieceType.Queen : fromData.ubQueen = fromData.ubQueen Or toPos.ToBitBoard
      Case PieceType.King : fromData.ubKing = fromData.ubKing Or toPos.ToBitBoard
    End Select

    If bFromNaught Then PieceData = fromData
  End Sub
  Private Sub ChangePiece(fromPos As PiecePos, toPos As PiecePos, Optional ByRef fromData As BoardData = Nothing)
    Dim bFromNaught As Boolean = False
    If fromData.IsEmpty Then fromData = PieceData : bFromNaught = True

    If CBool(fromData.ubWhite And fromPos.ToBitBoard) Then
      fromData.ubWhite = fromData.ubWhite And Not fromPos.ToBitBoard
      fromData.ubWhite = fromData.ubWhite Or toPos.ToBitBoard
    ElseIf CBool(fromData.ubBlack And fromPos.ToBitBoard) Then
      fromData.ubBlack = fromData.ubBlack And Not fromPos.ToBitBoard
      fromData.ubBlack = fromData.ubBlack Or toPos.ToBitBoard
    End If
    If CBool(fromData.ubPawn And fromPos.ToBitBoard) Then
      fromData.ubPawn = fromData.ubPawn And Not fromPos.ToBitBoard
      fromData.ubPawn = fromData.ubPawn Or toPos.ToBitBoard
    End If
    If CBool(fromData.ubRook And fromPos.ToBitBoard) Then
      fromData.ubRook = fromData.ubRook And Not fromPos.ToBitBoard
      fromData.ubRook = fromData.ubRook Or toPos.ToBitBoard
    End If
    If CBool(fromData.ubKnight And fromPos.ToBitBoard) Then
      fromData.ubKnight = fromData.ubKnight And Not fromPos.ToBitBoard
      fromData.ubKnight = fromData.ubKnight Or toPos.ToBitBoard
    End If
    If CBool(fromData.ubBishop And fromPos.ToBitBoard) Then
      fromData.ubBishop = fromData.ubBishop And Not fromPos.ToBitBoard
      fromData.ubBishop = fromData.ubBishop Or toPos.ToBitBoard
    End If
    If CBool(fromData.ubQueen And fromPos.ToBitBoard) Then
      fromData.ubQueen = fromData.ubQueen And Not fromPos.ToBitBoard
      fromData.ubQueen = fromData.ubQueen Or toPos.ToBitBoard
    End If
    If CBool(fromData.ubKing And fromPos.ToBitBoard) Then
      fromData.ubKing = fromData.ubKing And Not fromPos.ToBitBoard
      fromData.ubKing = fromData.ubKing Or toPos.ToBitBoard
    End If
    If CBool(fromData.ubStart And fromPos.ToBitBoard) Then fromData.ubStart = fromData.ubStart And Not fromPos.ToBitBoard

    If bFromNaught Then PieceData = fromData
  End Sub
  Private Sub RemovePiece(fromPos As PiecePos, Optional ByRef fromData As BoardData = Nothing)
    Dim bFromNaught As Boolean = False
    If fromData.IsEmpty Then fromData = PieceData : bFromNaught = True
    fromData.ubWhite = fromData.ubWhite And Not fromPos.ToBitBoard
    fromData.ubBlack = fromData.ubBlack And Not fromPos.ToBitBoard
    fromData.ubPawn = fromData.ubPawn And Not fromPos.ToBitBoard
    fromData.ubRook = fromData.ubRook And Not fromPos.ToBitBoard
    fromData.ubKnight = fromData.ubKnight And Not fromPos.ToBitBoard
    fromData.ubBishop = fromData.ubBishop And Not fromPos.ToBitBoard
    fromData.ubQueen = fromData.ubQueen And Not fromPos.ToBitBoard
    fromData.ubKing = fromData.ubKing And Not fromPos.ToBitBoard
    fromData.ubStart = fromData.ubStart And Not fromPos.ToBitBoard
    If bFromNaught Then PieceData = fromData
  End Sub
#End Region
#End Region
#Region "Data Functions"
#Region "Piece Data"
  Public Function GetPieceType(fromPos As PiecePos, Optional fromData As BoardData = Nothing) As PieceType
    If fromData.IsEmpty Then fromData = PieceData
    If CBool(fromPos.ToBitBoard And fromData.ubPawn) Then Return PieceType.Pawn
    If CBool(fromPos.ToBitBoard And fromData.ubRook) Then Return PieceType.Rook
    If CBool(fromPos.ToBitBoard And fromData.ubKnight) Then Return PieceType.Knight
    If CBool(fromPos.ToBitBoard And fromData.ubBishop) Then Return PieceType.Bishop
    If CBool(fromPos.ToBitBoard And fromData.ubQueen) Then Return PieceType.Queen
    If CBool(fromPos.ToBitBoard And fromData.ubKing) Then Return PieceType.King
    Return PieceType.Nothing
  End Function
  Public Function GetPieceColor(fromPos As PiecePos, Optional fromData As BoardData = Nothing) As Color
    If fromData.IsEmpty Then fromData = PieceData
    If CBool(fromPos.ToBitBoard And fromData.ubWhite) Then Return Color.White
    If CBool(fromPos.ToBitBoard And fromData.ubBlack) Then Return Color.Black
    Return Color.Transparent
  End Function
  Public Function GetPieceMoves(fromPos As PiecePos, Optional fromData As BoardData = Nothing) As PiecePos()
    If fromData.IsEmpty Then fromData = PieceData
    If CBool(fromPos.ToBitBoard And (fromData.ubBlack Or fromData.ubWhite)) Then
      Dim pType As PieceType = GetPieceType(fromPos, fromData)
      Dim isWhite As Boolean = GetPieceColor(fromPos, fromData) = Color.White
      Select Case pType
        Case PieceType.Pawn : Return BitFieldToPos(FieldPawn(fromPos, isWhite, fromData, True), True, fromPos, fromData)
        Case PieceType.Rook : Return BitFieldToPos(FieldRook(fromPos, isWhite, fromData, True), True, fromPos, fromData)
        Case PieceType.Knight : Return BitFieldToPos(FieldKnight(fromPos, isWhite, fromData, True), True, fromPos, fromData)
        Case PieceType.Bishop : Return BitFieldToPos(FieldBishop(fromPos, isWhite, fromData, True), True, fromPos, fromData)
        Case PieceType.Queen : Return BitFieldToPos(FieldQueen(fromPos, isWhite, fromData, True), True, fromPos, fromData)
        Case PieceType.King : Return BitFieldToPos(FieldKing(fromPos, isWhite, fromData, False), True, fromPos, fromData)
      End Select
    End If
    Return Nothing
  End Function
  Public Function isInCheck(whiteKing As Boolean, Optional fromData As BoardData = Nothing, Optional UseIgnore As Boolean = True) As Boolean
    If fromData.IsEmpty Then fromData = PieceData
    Return CBool(KingThreats(fromData, Not whiteKing, UseIgnore) And (fromData.ubKing And IIf(whiteKing, fromData.ubWhite, fromData.ubBlack)))
  End Function
  Private Function leadsToCheck(fromPoint As PiecePos, toPoint As PiecePos, bData As BoardData) As Boolean
    Dim isWhite As Boolean = CBool(fromPoint.ToBitBoard And bData.ubWhite)
    Dim pType As PieceType = GetPieceType(fromPoint, bData)
    RemovePiece(toPoint, bData)
    ChangePiece(fromPoint, toPoint, bData)
    Return isInCheck(isWhite, bData, False)
  End Function
  Private Function KingThreats(bData As BoardData, fromWhite As Boolean, IgnoreKingSafety As Boolean) As UInt64
    Dim uRet As UInt64 = 0
    If fromWhite Then
      For I As Integer = 0 To 63
        Dim iPiece As UInt64 = 1UL << I
        If CBool(iPiece And bData.ubWhite) Then
          If CBool(iPiece And bData.ubKing) Then
            uRet = uRet Or FieldKing(New PiecePos(iPiece), fromWhite, bData, IgnoreKingSafety)
          ElseIf CBool(iPiece And bData.ubQueen) Then
            uRet = uRet Or FieldQueen(New PiecePos(iPiece), fromWhite, bData, False)
          ElseIf CBool(iPiece And bData.ubBishop) Then
            uRet = uRet Or FieldBishop(New PiecePos(iPiece), fromWhite, bData, False)
          ElseIf CBool(iPiece And bData.ubKnight) Then
            uRet = uRet Or FieldKnight(New PiecePos(iPiece), fromWhite, bData, False)
          ElseIf CBool(iPiece And bData.ubRook) Then
            uRet = uRet Or FieldRook(New PiecePos(iPiece), fromWhite, bData, False)
          ElseIf CBool(iPiece And bData.ubPawn) Then
            uRet = uRet Or FieldPawn_Attack(New PiecePos(iPiece), fromWhite, bData, False)
          End If
        End If
      Next
    Else
      For I As Integer = 0 To 63
        Dim iPiece As UInt64 = 1UL << I
        If CBool(iPiece And bData.ubBlack) Then
          If CBool(iPiece And bData.ubKing) Then
            uRet = uRet Or FieldKing(New PiecePos(iPiece), fromWhite, bData, IgnoreKingSafety)
          ElseIf CBool(iPiece And bData.ubQueen) Then
            uRet = uRet Or FieldQueen(New PiecePos(iPiece), fromWhite, bData, False)
          ElseIf CBool(iPiece And bData.ubBishop) Then
            uRet = uRet Or FieldBishop(New PiecePos(iPiece), fromWhite, bData, False)
          ElseIf CBool(iPiece And bData.ubKnight) Then
            uRet = uRet Or FieldKnight(New PiecePos(iPiece), fromWhite, bData, False)
          ElseIf CBool(iPiece And bData.ubRook) Then
            uRet = uRet Or FieldRook(New PiecePos(iPiece), fromWhite, bData, False)
          ElseIf CBool(iPiece And bData.ubPawn) Then
            uRet = uRet Or FieldPawn_Attack(New PiecePos(iPiece), fromWhite, bData, False)
          End If
        End If
      Next
    End If
    Return uRet
  End Function
#End Region
  Public Function BoardToString() As String
    Dim sBoard As String = String.Empty
    sBoard &= "WK:" & (New PiecePos(PieceData.ubKing And PieceData.ubWhite)).ToString & IIf(PieceData.ubStart And (PieceData.ubKing And PieceData.ubWhite), "u", "") & ","
    sBoard &= "BK:" & (New PiecePos(PieceData.ubKing And PieceData.ubBlack)).ToString & IIf(PieceData.ubStart And (PieceData.ubKing And PieceData.ubBlack), "u", "") & ","
    If (PieceData.ubQueen And PieceData.ubWhite) > 0 Then sBoard &= "WQ:" & PosArrayToString(BitFieldToPos(PieceData.ubQueen And PieceData.ubWhite, False)) & ","
    If (PieceData.ubQueen And PieceData.ubBlack) > 0 Then sBoard &= "BQ:" & PosArrayToString(BitFieldToPos(PieceData.ubQueen And PieceData.ubBlack, False)) & ","
    If (PieceData.ubBishop And PieceData.ubWhite) > 0 Then sBoard &= "WB:" & PosArrayToString(BitFieldToPos(PieceData.ubBishop And PieceData.ubWhite, False)) & ","
    If (PieceData.ubBishop And PieceData.ubBlack) > 0 Then sBoard &= "BB:" & PosArrayToString(BitFieldToPos(PieceData.ubBishop And PieceData.ubBlack, False)) & ","
    If (PieceData.ubKnight And PieceData.ubWhite) > 0 Then sBoard &= "WN:" & PosArrayToString(BitFieldToPos(PieceData.ubKnight And PieceData.ubWhite, False)) & ","
    If (PieceData.ubKnight And PieceData.ubBlack) > 0 Then sBoard &= "BN:" & PosArrayToString(BitFieldToPos(PieceData.ubKnight And PieceData.ubBlack, False)) & ","
    If (PieceData.ubRook And PieceData.ubWhite) > 0 Then sBoard &= "WR:" & PosArrayToString(BitFieldToPos(PieceData.ubRook And PieceData.ubWhite, False)) & IIf(PieceData.ubStart And (PieceData.ubRook And PieceData.ubWhite), "u", "") & ","
    If (PieceData.ubRook And PieceData.ubBlack) > 0 Then sBoard &= "BR:" & PosArrayToString(BitFieldToPos(PieceData.ubRook And PieceData.ubBlack, False)) & IIf(PieceData.ubStart And (PieceData.ubRook And PieceData.ubBlack), "u", "") & ","
    If (PieceData.ubPawn And PieceData.ubWhite) > 0 Then sBoard &= "Wp:" & PosArrayToString(BitFieldToPos(PieceData.ubPawn And PieceData.ubWhite, False)) & ","
    If (PieceData.ubPawn And PieceData.ubBlack) > 0 Then sBoard &= "Bp:" & PosArrayToString(BitFieldToPos(PieceData.ubPawn And PieceData.ubBlack, False)) & ","
    If PieceData.ubWEnP > 0 Then sBoard &= "We:" & PosArrayToString(BitFieldToPos(PieceData.ubWEnP, False)) & ","
    If PieceData.ubBEnP > 0 Then sBoard &= "Be:" & PosArrayToString(BitFieldToPos(PieceData.ubBEnP, False)) & ","
    Return sBoard.Substring(0, sBoard.Length - 1)
  End Function
  Private Function PosArrayToString(pos As PiecePos()) As String
    If pos.Length = 1 Then
      Return pos(0).ToString
    Else
      Dim sRet As String = String.Empty
      For I As Integer = 0 To pos.Length - 1
        sRet &= pos(I).ToString & "+"
      Next
      Return sRet.Substring(0, sRet.Length - 1)
    End If
  End Function
  Public Sub StringToBoard(sBoard As String)
    Dim bTmp As BoardData

    Dim sPieceDescriptions() As String = sBoard.Split(","c)
    For Each sDescr As String In sPieceDescriptions
      Dim uPoints As UInt64
      Dim uStart As UInt64
      If sDescr.Contains("+") Then
        Dim sPoints() As String
        sPoints = sDescr.Substring(3).Split("+")
        uPoints = 0
        For I As Integer = 0 To sPoints.Length - 1
          If sPoints(I).Length > 2 Then
            uPoints = uPoints Or (New PiecePos(sPoints(I).Substring(0, 2))).ToBitBoard
            If sPoints(I).Substring(2, 1) = "u" Then uStart = uStart Or (New PiecePos(sPoints(I).Substring(0, 2))).ToBitBoard
          Else
            uPoints = uPoints Or (New PiecePos(sPoints(I))).ToBitBoard
          End If
        Next
      Else
        uPoints = (New PiecePos(sDescr.Substring(3, 2))).ToBitBoard
        If sDescr.Length > 5 AndAlso sDescr.Substring(5, 1) = "u" Then uStart = uStart Or (New PiecePos(sDescr.Substring(3, 2))).ToBitBoard
      End If
      Select Case sDescr.Trim.Substring(0, 2)
        Case "WK"
          bTmp.ubWhite = bTmp.ubWhite Or uPoints
          bTmp.ubKing = bTmp.ubKing Or uPoints
        Case "BK"
          bTmp.ubBlack = bTmp.ubBlack Or uPoints
          bTmp.ubKing = bTmp.ubKing Or uPoints
        Case "WQ"
          bTmp.ubWhite = bTmp.ubWhite Or uPoints
          bTmp.ubQueen = bTmp.ubQueen Or uPoints
        Case "BQ"
          bTmp.ubBlack = bTmp.ubBlack Or uPoints
          bTmp.ubQueen = bTmp.ubQueen Or uPoints
        Case "WB"
          bTmp.ubWhite = bTmp.ubWhite Or uPoints
          bTmp.ubBishop = bTmp.ubBishop Or uPoints
        Case "BB"
          bTmp.ubBlack = bTmp.ubBlack Or uPoints
          bTmp.ubBishop = bTmp.ubBishop Or uPoints
        Case "WN"
          bTmp.ubWhite = bTmp.ubWhite Or uPoints
          bTmp.ubKnight = bTmp.ubKnight Or uPoints
        Case "BN"
          bTmp.ubBlack = bTmp.ubBlack Or uPoints
          bTmp.ubKnight = bTmp.ubKnight Or uPoints
        Case "WR"
          bTmp.ubWhite = bTmp.ubWhite Or uPoints
          bTmp.ubRook = bTmp.ubRook Or uPoints
        Case "BR"
          bTmp.ubBlack = bTmp.ubBlack Or uPoints
          bTmp.ubRook = bTmp.ubRook Or uPoints
        Case "Wp"
          bTmp.ubWhite = bTmp.ubWhite Or uPoints
          bTmp.ubPawn = bTmp.ubPawn Or uPoints
        Case "Bp"
          bTmp.ubBlack = bTmp.ubBlack Or uPoints
          bTmp.ubPawn = bTmp.ubPawn Or uPoints
        Case "We"
          bTmp.ubWEnP = bTmp.ubWEnP Or uPoints
        Case "Be"
          bTmp.ubBEnP = bTmp.ubBEnP Or uPoints
      End Select
      If uStart > 0 Then bTmp.ubStart = bTmp.ubStart Or uStart
    Next

    PieceData = bTmp
  End Sub
  Private Function BitBoardToString(bb As UInt64) As String
    Dim bA As New BitArray(BitConverter.GetBytes(bb))
    Dim sRet As String = String.Empty
    For I As Integer = 0 To bA.Length - 1
      sRet &= IIf(bA(I), "1", "0")
      If I Mod 8 = 7 Then sRet &= vbNewLine
    Next
    Return StrReverse(sRet).Replace(vbLf & vbCr, vbCrLf)
  End Function
  Private Function BitFieldToPos(fromField As UInt64, checkForCheck As Boolean, Optional fromPos As PiecePos = Nothing, Optional fromData As BoardData = Nothing) As PiecePos()
    If fromData.IsEmpty Then fromData = PieceData
    Dim pRets As New List(Of PiecePos)
    For X As Integer = 1 To 8
      For Y As Integer = 1 To 8
        Dim pRet As New PiecePos(Y, X)
        If CBool(pRet.ToBitBoard And fromField) Then
          If checkForCheck Then
            If Not leadsToCheck(fromPos, pRet, fromData) Then
              pRets.Add(pRet)
            End If
          Else
            pRets.Add(pRet)
          End If
        End If
      Next
    Next
    Return pRets.ToArray
  End Function
#End Region
#Region "Continue Functions"
  Public Sub SaveGame(Path As String, isWhite As Boolean, myTurn As Boolean)
    Using ioSave As New IO.FileStream(Path, IO.FileMode.Create, IO.FileAccess.Write, IO.FileShare.None)
      Using ioGame As New IO.BinaryWriter(ioSave)
        ioGame.Write("CHESS")
        ioGame.Write(isWhite)
        ioGame.Write(myTurn)
        ioGame.Write(PieceData.ubWhite)
        ioGame.Write(PieceData.ubBlack)
        ioGame.Write(PieceData.ubKing)
        ioGame.Write(PieceData.ubQueen)
        ioGame.Write(PieceData.ubBishop)
        ioGame.Write(PieceData.ubKnight)
        ioGame.Write(PieceData.ubRook)
        ioGame.Write(PieceData.ubPawn)
        ioGame.Write(PieceData.ubStart)
        ioGame.Write(PieceData.ubWEnP)
        ioGame.Write(PieceData.ubBEnP)
        ioGame.Flush()
      End Using
    End Using
  End Sub
  Public Function LoadGame(Path As String, ByRef isWhite As Boolean, ByRef myTurn As Boolean) As Boolean
    Using ioSave As New IO.FileStream(Path, IO.FileMode.Open, IO.FileAccess.Read, IO.FileShare.Read)
      Using ioGame As New IO.BinaryReader(ioSave)
        If ioGame.ReadString() = "CHESS" Then
          isWhite = ioGame.ReadBoolean
          myTurn = ioGame.ReadBoolean
          PieceData.ubWhite = ioGame.ReadUInt64
          PieceData.ubBlack = ioGame.ReadUInt64
          PieceData.ubKing = ioGame.ReadUInt64
          PieceData.ubQueen = ioGame.ReadUInt64
          PieceData.ubBishop = ioGame.ReadUInt64
          PieceData.ubKnight = ioGame.ReadUInt64
          PieceData.ubRook = ioGame.ReadUInt64
          PieceData.ubPawn = ioGame.ReadUInt64
          PieceData.ubStart = ioGame.ReadUInt64
          PieceData.ubWEnP = ioGame.ReadUInt64
          PieceData.ubBEnP = ioGame.ReadUInt64
        End If
      End Using
    End Using
    Return isWhite
  End Function
#End Region
End Class