﻿Public Class PcxIcon
  Private C_Image As Drawing.Image
  Public ReadOnly Property Image As Drawing.Image
    Get
      Return C_Image
    End Get
  End Property
  Public Sub New(img As Drawing.Image)
    C_Image = img
  End Sub
End Class
Public Class PcxFileParser
  Private C_Image As Drawing.Image
  Private C_Icons() As PcxIcon
  Public ReadOnly Property AllIcons As PcxIcon()
    Get
      Return C_Icons
    End Get
  End Property
  Public ReadOnly Property FullImage As Drawing.Image
    Get
      Return C_Image
    End Get
  End Property
  Public Sub New(filePath As String)
    Dim PCX As FreeImageAPI.FIBITMAP = FreeImageAPI.FreeImage.Load(FreeImageAPI.FREE_IMAGE_FORMAT.FIF_PCX, filePath, FreeImageAPI.FREE_IMAGE_LOAD_FLAGS.DEFAULT)
    ParsePCX(PCX)
    PCX.SetNull()
  End Sub
  Public Sub New(pcxFileStream As System.IO.Stream)
    Dim PCX As FreeImageAPI.FIBITMAP = FreeImageAPI.FreeImage.LoadFromStream(pcxFileStream, FreeImageAPI.FREE_IMAGE_LOAD_FLAGS.DEFAULT, FreeImageAPI.FREE_IMAGE_FORMAT.FIF_PCX)
    ParsePCX(PCX)
    PCX.SetNull()
  End Sub
  Private Sub ParsePCX(PCX As FreeImageAPI.FIBITMAP)
    Dim bmpStream As New IO.MemoryStream
    FreeImageAPI.FreeImage.SaveToStream(PCX, bmpStream, FreeImageAPI.FREE_IMAGE_FORMAT.FIF_BMP)
    C_Image = Image.FromStream(bmpStream)
    bmpStream.Close()
    Dim iWidth As Integer = C_Image.Width / 28 - 1
    Dim iHeight As Integer = C_Image.Height / 14 - 1
    ReDim C_Icons((iWidth + 1) * (iHeight + 1))
    Dim I As Integer = 0
    For X As Integer = 0 To iWidth
      For Y As Integer = 0 To iHeight
        Dim tIMG As Image = New System.Drawing.Bitmap(28, 14)
        Dim g As Graphics = Graphics.FromImage(tIMG)
        g.InterpolationMode = Drawing2D.InterpolationMode.HighQualityBicubic
        g.DrawImage(C_Image, -(X * 28), -(Y * 14), C_Image.Width, C_Image.Height)
        g.Dispose()
        Dim cPCX As New PcxIcon(tIMG)
        C_Icons(I) = cPCX
        I += 1
      Next Y
    Next X
  End Sub
End Class
Public Class LargePcxFileParser
  Private C_Image As Drawing.Image
  Private C_Icons() As PcxIcon
  Public ReadOnly Property AllIcons As PcxIcon()
    Get
      Return C_Icons
    End Get
  End Property
  Public ReadOnly Property FullImage As Drawing.Image
    Get
      Return C_Image
    End Get
  End Property
  Public Sub New(filePath As String)
    Dim PCX As FreeImageAPI.FIBITMAP = FreeImageAPI.FreeImage.Load(FreeImageAPI.FREE_IMAGE_FORMAT.FIF_PCX, filePath, FreeImageAPI.FREE_IMAGE_LOAD_FLAGS.DEFAULT)
    ParsePCX(PCX)
    PCX.SetNull()
  End Sub
  Public Sub New(pcxFileStream As System.IO.Stream)
    Dim PCX As FreeImageAPI.FIBITMAP = FreeImageAPI.FreeImage.LoadFromStream(pcxFileStream, FreeImageAPI.FREE_IMAGE_LOAD_FLAGS.DEFAULT, FreeImageAPI.FREE_IMAGE_FORMAT.FIF_PCX)
    ParsePCX(PCX)
    PCX.SetNull()
  End Sub
  Private Sub ParsePCX(PCX As FreeImageAPI.FIBITMAP)
    Dim bmpStream As New IO.MemoryStream
    FreeImageAPI.FreeImage.SaveToStream(PCX, bmpStream, FreeImageAPI.FREE_IMAGE_FORMAT.FIF_BMP)
    C_Image = Image.FromStream(bmpStream)
    bmpStream.Close()
    Dim iWidth As Integer = C_Image.Width / 64 - 1
    Dim iHeight As Integer = C_Image.Height / 38 - 1
    ReDim C_Icons((iWidth + 1) * (iHeight + 1))
    Dim I As Integer = 0
    For X As Integer = 0 To iWidth
      For Y As Integer = 0 To iHeight
        Dim tIMG As Image = New System.Drawing.Bitmap(64, 38)
        Dim g As Graphics = Graphics.FromImage(tIMG)
        g.InterpolationMode = Drawing2D.InterpolationMode.HighQualityBicubic
        g.DrawImage(C_Image, -(X * 64), -(Y * 38), C_Image.Width, C_Image.Height)
        g.Dispose()
        Dim cPCX As New PcxIcon(tIMG)
        C_Icons(I) = cPCX
        I += 1
      Next Y
    Next X
  End Sub
End Class
Public Class LargePngFileParser
  Private C_Image As Drawing.Image
  Private C_Icons() As Image
  Public ReadOnly Property AllIcons As Image()
    Get
      Return C_Icons
    End Get
  End Property
  Public ReadOnly Property FullImage As Drawing.Image
    Get
      Return C_Image
    End Get
  End Property
  Public Sub New(filePath As String)
    Using PNG As Image = Image.FromFile(filePath)
      ParsePNG(PNG)
    End Using
  End Sub
  Public Sub New(pngFileStream As System.IO.Stream)
    Using PNG As Image = Image.FromStream(pngFileStream)
      ParsePNG(PNG)
    End Using
  End Sub
  Private Sub ParsePNG(PNG As Image)
    C_Image = PNG.Clone
    Dim iWidth As Integer = C_Image.Width / 64 - 1
    Dim iHeight As Integer = C_Image.Height / 38 - 1
    ReDim C_Icons((iWidth + 1) * (iHeight + 1))
    Dim I As Integer = 0
    For X As Integer = 0 To iWidth
      For Y As Integer = 0 To iHeight
        Dim tIMG As Bitmap = New System.Drawing.Bitmap(64, 38)
        Dim g As Graphics = Graphics.FromImage(tIMG)
        g.Clear(Color.Transparent)
        g.DrawImage(C_Image, -(X * 64), -(Y * 38), C_Image.Width, C_Image.Height)
        g.Dispose()
        C_Icons(I) = tIMG.Clone
        I += 1
      Next Y
    Next X
  End Sub
End Class
Public Module modIconHandler
  Public ReadOnly Property Icons_Temp() As String
    Get
      Static sIcons As String
      If String.IsNullOrEmpty(sIcons) Then sIcons = AppTemp & "\icons\"
      Return sIcons
    End Get
  End Property
  Public ReadOnly Property Icons_Store() As String
    Get
      Static sIcons As String
      If String.IsNullOrEmpty(sIcons) Then sIcons = AppStore & "\icons\"
      Return sIcons
    End Get
  End Property
  Public Sub ParseAllIcons()
    If Not My.Computer.FileSystem.DirectoryExists(Icons_Temp) Then My.Computer.FileSystem.CreateDirectory(Icons_Temp)
    If Not My.Computer.FileSystem.DirectoryExists(Icons_Store) Then My.Computer.FileSystem.CreateDirectory(Icons_Store)
    ParseStandardIcons()
    Dim bW3 As Boolean = ParseWAR3Icons()
    If My.Computer.FileSystem.FileExists(Icons_Store & "icons_STAR.bni") AndAlso Not My.Computer.FileSystem.DirectoryExists(Icons_Store & "icons_STAR") Then ParseIconsSTAR()
    If My.Computer.FileSystem.FileExists(Icons_Store & "icons-WAR3.bni") AndAlso bW3 Then ParseIconsWAR3()
  End Sub
  Private Function CheckDir(Directory As String, ImageCount As Integer) As Boolean
    Dim bDo As Boolean = False
    If My.Computer.FileSystem.DirectoryExists(Directory) Then
      For I As Integer = 0 To ImageCount - 1
        If Not My.Computer.FileSystem.FileExists(Directory & "\" & Trim(CStr(I)) & ".png") Then
          bDo = True
          Exit For
        End If
      Next I
    Else
      bDo = True
    End If
    Return bDo
  End Function
  Private Sub ParseStandardIcons()
    ParseIconsDefault()
    If CheckDir(Icons_Temp & "DRTL", 12) Then SavePCX("DRTL", My.Resources.DRTL)
    If CheckDir(Icons_Temp & "DSHR", 1) Then SavePCX("DSHR", My.Resources.DSHR)
    If CheckDir(Icons_Temp & "STAR", 72) Then SavePCX("STAR", My.Resources.STAR)
    If CheckDir(Icons_Temp & "D2DV", 420) Then SavePCX("D2DV", My.Resources.D2DV)
    If CheckDir(Icons_Temp & "D2XP", 714) Then SavePCX("D2XP", My.Resources.D2XP)

    If CheckDir(Icons_Temp & "Full\L_DRTL", 12) Then SaveLargePCX("L_DRTL", My.Resources.L_DRTL)
    If CheckDir(Icons_Temp & "Full\L_DSHR", 2) Then SaveLargePNG("L_DSHR", My.Resources.L_DSHR)
    If CheckDir(Icons_Temp & "Full\L_STAR", 16) Then SaveLargePNG("L_STAR", My.Resources.L_STAR)
    If CheckDir(Icons_Temp & "Full\L_JSTR", 2) Then SaveLargePNG("L_JSTR", My.Resources.L_JSTR)
    If CheckDir(Icons_Temp & "Full\L_W2BN", 24) Then SaveLargePNG("L_W2BN", My.Resources.L_W2BN)

    If CheckDir(Icons_Temp & "Full\L_D2DV", 420) Then SaveLargePCX("L_D2DV", My.Resources.L_D2DV)
    If CheckDir(Icons_Temp & "Full\L_D2XP", 714) Then SaveLargePCX("L_D2XP", My.Resources.L_D2XP)

    If CheckDir(Icons_Temp & "W2BN", 34) Then SavePCX("W2BN", My.Resources.W2BN)
  End Sub
  Private Function ParseWAR3Icons() As Boolean
    Dim bLoad As Boolean = False
    If My.Computer.FileSystem.DirectoryExists(Icons_Temp & "WAR3") Then
      If My.Computer.FileSystem.FileExists(Icons_Temp & "WAR3\iconindex_clan.txt") Then
        Dim sClan() As String = Split(My.Computer.FileSystem.ReadAllText(Icons_Temp & "WAR3\iconindex_clan.txt"), vbNewLine)
        For Each sIcon As String In sClan
          If sIcon.Length > 2 Then
            If Not My.Computer.FileSystem.FileExists(Icons_Temp & "WAR3\" & sIcon.Substring(sIcon.LastIndexOf("\") + 1) & ".png") Then
              bLoad = True
              Exit For
            End If
          End If
        Next
      End If

      If My.Computer.FileSystem.FileExists(Icons_Temp & "WAR3\iconindex_bnet.txt") Then
        Dim sClan() As String = Split(My.Computer.FileSystem.ReadAllText(Icons_Temp & "WAR3\iconindex_bnet.txt"), vbNewLine)
        For Each sIcon As String In sClan
          If sIcon.Length > 2 Then
            If Not My.Computer.FileSystem.FileExists(Icons_Temp & "WAR3\" & sIcon.Substring(sIcon.LastIndexOf("\") + 1) & ".png") Then
              bLoad = True
              Exit For
            End If
          End If
        Next
      End If

      If Not bLoad AndAlso My.Computer.FileSystem.FileExists(Icons_Temp & "WAR3\iconindex_def.txt") Then
        Dim sDef() As String = Split(My.Computer.FileSystem.ReadAllText(Icons_Temp & "WAR3\iconindex_def.txt"), vbNewLine)
        For Each sIcon As String In sDef
          If sIcon.Length > 2 Then
            If Not My.Computer.FileSystem.FileExists(Icons_Temp & "WAR3\" & sIcon.Substring(sIcon.LastIndexOf("\") + 1) & ".png") Then
              bLoad = True
              Exit For
            End If
          End If
        Next
      End If
      If Not bLoad AndAlso My.Computer.FileSystem.FileExists(Icons_Temp & "WAR3\iconindex_def2.txt") Then
        Dim sDef2() As String = Split(My.Computer.FileSystem.ReadAllText(Icons_Temp & "WAR3\iconindex_def2.txt"), vbNewLine)
        For Each sIcon As String In sDef2
          If sIcon.Length > 2 Then
            If Not My.Computer.FileSystem.FileExists(Icons_Temp & "WAR3\" & sIcon.Substring(sIcon.LastIndexOf("\") + 1) & ".png") Then
              bLoad = True
              Exit For
            End If
          End If
        Next
      End If
    Else
      bLoad = True
    End If
    If (Not My.Computer.FileSystem.FileExists(Icons_Store & "WAR3.bni")) OrElse (Not My.Computer.FileSystem.GetFileInfo(Icons_Store & "WAR3.bni").Length = My.Resources.WAR3.Length) Then bLoad = True
    If bLoad Then SaveBLP("WAR3", "WAR3", My.Resources.WAR3)
    Return bLoad
  End Function
  Public Sub ParseIconsDefault()
    Dim bLoad As Boolean = False
    If My.Computer.FileSystem.DirectoryExists(Icons_Temp & "icons") Then
      If My.Computer.FileSystem.FileExists(Icons_Temp & "icons\DRTL.png") AndAlso
         My.Computer.FileSystem.FileExists(Icons_Temp & "icons\DSHR.png") AndAlso
         My.Computer.FileSystem.FileExists(Icons_Temp & "icons\D2DV.png") AndAlso
         My.Computer.FileSystem.FileExists(Icons_Temp & "icons\D2XP.png") AndAlso
         My.Computer.FileSystem.FileExists(Icons_Temp & "icons\STAR.png") AndAlso
         My.Computer.FileSystem.FileExists(Icons_Temp & "icons\SEXP.png") AndAlso
         My.Computer.FileSystem.FileExists(Icons_Temp & "icons\JSTR.png") AndAlso
         My.Computer.FileSystem.FileExists(Icons_Temp & "icons\SSHR.png") AndAlso
         My.Computer.FileSystem.FileExists(Icons_Temp & "icons\W2BN.png") AndAlso
         My.Computer.FileSystem.FileExists(Icons_Temp & "icons\W3XP.png") AndAlso
         My.Computer.FileSystem.FileExists(Icons_Temp & "icons\CHAT.png") AndAlso
         My.Computer.FileSystem.FileExists(Icons_Temp & "icons\FLAG_1.png") AndAlso
         My.Computer.FileSystem.FileExists(Icons_Temp & "icons\FLAG_2.png") AndAlso
         My.Computer.FileSystem.FileExists(Icons_Temp & "icons\FLAG_20.png") Then
        bLoad = False
      Else
        bLoad = True
      End If
    Else
      bLoad = True
    End If
    If bLoad Then SaveBNI("icons")
  End Sub
  Public Sub ParseIconsSTAR()
    SaveBNI("icons_STAR")
  End Sub
  Public Sub ParseIconsWAR3()
    SaveBLP("icons-WAR3", "WAR3")
  End Sub
  Private Sub SavePCX(FileName As String, ByRef Resource() As Byte)
    Dim bDoIt As Boolean = False
    If Not My.Computer.FileSystem.DirectoryExists(Icons_Temp & FileName) Then My.Computer.FileSystem.CreateDirectory(Icons_Temp & FileName)
    LoadIcons(FileName & ".pcx", Resource)
    Dim parse As New PcxFileParser(Icons_Store & FileName & ".pcx")
    Dim Icons() As PcxIcon = parse.AllIcons
    For I As Integer = 0 To Icons.Length - 1
      Dim Icon As PcxIcon = Icons(I)
      If Icon IsNot Nothing Then Icon.Image.Save(Icons_Temp & FileName & "\" & I & ".png", Imaging.ImageFormat.Png)
    Next I
  End Sub
  Private Sub SaveLargePCX(FileName As String, ByRef Resource() As Byte)
    Dim bDoIt As Boolean = False
    If Not My.Computer.FileSystem.DirectoryExists(Icons_Temp & "Full") Then My.Computer.FileSystem.CreateDirectory(Icons_Temp & "Full")
    If Not My.Computer.FileSystem.DirectoryExists(Icons_Temp & "Full\" & FileName) Then My.Computer.FileSystem.CreateDirectory(Icons_Temp & "Full\" & FileName)
    LoadIcons(FileName & ".pcx", Resource)
    Dim parse As New LargePcxFileParser(Icons_Store & FileName & ".pcx")
    Dim Icons() As PcxIcon = parse.AllIcons
    For I As Integer = 0 To Icons.Length - 1
      Dim Icon As PcxIcon = Icons(I)
      If Icon IsNot Nothing Then Icon.Image.Save(Icons_Temp & "Full\" & FileName & "\" & I & ".png", Imaging.ImageFormat.Png)
    Next I
  End Sub
  Private Sub SaveLargePNG(FileName As String, ByRef Resource As Bitmap)
    Dim bDoIt As Boolean = False
    If Not My.Computer.FileSystem.DirectoryExists(Icons_Temp & "Full") Then My.Computer.FileSystem.CreateDirectory(Icons_Temp & "Full")
    If Not My.Computer.FileSystem.DirectoryExists(Icons_Temp & "Full\" & FileName) Then My.Computer.FileSystem.CreateDirectory(Icons_Temp & "Full\" & FileName)
    Resource.Save(Icons_Store & FileName & ".png", Imaging.ImageFormat.Png)

    Dim parse As New LargePngFileParser(Icons_Store & FileName & ".png")
    Dim Icons() As Image = parse.AllIcons
    For I As Integer = 0 To Icons.Length - 1
      Dim Icon As Image = Icons(I)
      If Icon IsNot Nothing Then Icon.Save(Icons_Temp & "Full\" & FileName & "\" & I & ".png", Imaging.ImageFormat.Png)
    Next I
  End Sub
  Private Sub SaveBNI(FileName As String)
    If My.Computer.FileSystem.FileExists(Icons_Store & FileName & ".bni") Then
      If Not My.Computer.FileSystem.DirectoryExists(Icons_Temp & FileName) Then My.Computer.FileSystem.CreateDirectory(Icons_Temp & FileName)
      Dim parse As New BniFileParser(Icons_Store & FileName & ".bni")
      Dim Icons() As BniIcon = parse.AllIcons
      parse.Dispose()
      For Each Icon As BniIcon In Icons
        If Icon.SoftwareProductCodes.Length > 0 Then
          If Not My.Computer.FileSystem.FileExists(Icons_Temp & FileName & "\" & Icon.SoftwareProductCodes(0) & ".png") Then Icon.Image.Save(Icons_Temp & FileName & "\" & Icon.SoftwareProductCodes(0) & ".png", Imaging.ImageFormat.Png)
        Else
          If Not My.Computer.FileSystem.FileExists(Icons_Temp & FileName & "\FLAG_" & Hex(Icon.UserFlags) & ".png") Then Icon.Image.Save(Icons_Temp & FileName & "\FLAG_" & Hex(Icon.UserFlags) & ".png", Imaging.ImageFormat.Png)
        End If
      Next Icon
    End If
  End Sub
  Private Sub SaveBLP(FileName As String, Directory As String, Optional ByRef Resource() As Byte = Nothing)
    If Not My.Computer.FileSystem.DirectoryExists(Icons_Temp & Directory) Then My.Computer.FileSystem.CreateDirectory(Icons_Temp & Directory)
    If Resource IsNot Nothing Then LoadIcons(FileName & ".bni", Resource)
    Dim Archive As New MPQ(Icons_Store & FileName & ".bni")
    Dim sList As String = Archive.GetListFile
    If String.IsNullOrEmpty(sList) Then Exit Sub
    Dim sFile() As String = Split(sList, vbNewLine)
    For Each ListItem As String In sFile
      If String.IsNullOrEmpty(ListItem) Then Continue For
      If ListItem.Substring(0, 1) = "(" Then Continue For
      If ListItem.Substring(ListItem.Length - 3) = "blp" Then
        Dim bData() As Byte = Archive.OpenFile(ListItem)
        Dim BLPImg As Image = New Bitmap(28, 14)
        Dim BLPSrc As Image = ParseBLP(bData, 0)
        Dim ImagePath As String = Icons_Temp & Directory & "\" & ListItem.Substring(ListItem.LastIndexOf("\") + 1) & ".png"
        If BLPSrc.Size.Width = 64 And BLPSrc.Size.Height = 64 Then
          If Not My.Computer.FileSystem.DirectoryExists(Icons_Temp & "Full\" & Directory) Then My.Computer.FileSystem.CreateDirectory(Icons_Temp & "Full\" & Directory)
          Dim FullPath As String = Icons_Temp & "Full\" & Directory & "\" & ListItem.Substring(ListItem.LastIndexOf("\") + 1) & ".png"
          Using g As Graphics = Graphics.FromImage(BLPImg)
            g.InterpolationMode = Drawing2D.InterpolationMode.HighQualityBicubic
            g.DrawImage(BLPSrc, 0, -CInt(Math.Floor(((BLPImg.Height / BLPImg.Width) * BLPSrc.Height) / 4)), BLPImg.Width, BLPImg.Width)
          End Using
          BLPImg.Save(ImagePath, Imaging.ImageFormat.Png)
          Dim blpFull As Image = New Bitmap(64, 38)
          Using g As Graphics = Graphics.FromImage(blpFull)
            g.InterpolationMode = Drawing2D.InterpolationMode.HighQualityBicubic
            g.DrawImage(BLPSrc, 0, -CInt(Math.Floor(((blpFull.Height / blpFull.Width) * BLPSrc.Height) / 4) + 4), blpFull.Width, blpFull.Width)
          End Using
          blpFull.Save(FullPath, Imaging.ImageFormat.Png)
        Else
          BLPSrc.Save(ImagePath, Imaging.ImageFormat.Png)
        End If
      ElseIf ListItem.Substring(ListItem.Length - 3) = "txt" Then
        Archive.SaveToPath(ListItem, Icons_Temp & Directory)
      Else
        Debug.Print("Unknown file in BLP: " & ListItem)
      End If
    Next
  End Sub
  Private Function LoadIcons(FileName As String, ByRef Resource() As Byte) As Boolean
    If (Not My.Computer.FileSystem.FileExists(Icons_Store & FileName)) OrElse (Not My.Computer.FileSystem.GetFileInfo(Icons_Store & FileName).Length = Resource.Length) Then
      If Not My.Computer.FileSystem.DirectoryExists(Icons_Store) Then My.Computer.FileSystem.CreateDirectory(Icons_Store)
      My.Computer.FileSystem.WriteAllBytes(Icons_Store & FileName, Resource, False)
      Return True
    End If
    Return False
  End Function
  Private Function ParseBLP(bFile() As Byte, Item As Byte) As Image
    Dim cBG As Color = Color.Transparent
    Dim dFile As New DataReader(bFile)
    Dim ID As String = System.Text.Encoding.GetEncoding(LATIN_1).GetString(dFile.ReadByteArray(4))
    If ID = "BLP1" Then
      Dim Compression As UInt32 = dFile.ReadUInt32
      Dim Flags As UInt32 = dFile.ReadUInt32
      Dim Width As Integer = dFile.ReadUInt32
      Dim Height As Integer = dFile.ReadUInt32
      Dim PictureType As UInt32 = dFile.ReadUInt32
      Dim PictureSubType As UInt32 = dFile.ReadUInt32
      Dim MiniMapOffset(15) As UInt32
      MiniMapOffset = dFile.ReadUInt32Array(16)
      Dim MiniMapSize(15) As UInt32
      MiniMapSize = dFile.ReadUInt32Array(16)
      If Compression = 0 Then
        Dim HeaderSize As UInt32 = dFile.ReadUInt32
        Dim Header() As Byte = dFile.ReadByteArray(HeaderSize)
        Dim JPGFile(HeaderSize + MiniMapSize(Item)) As Byte
        Array.Copy(Header, 0, JPGFile, 0, HeaderSize)
        Array.Copy(bFile, MiniMapOffset(Item), JPGFile, HeaderSize, MiniMapSize(Item))
        Dim iStream As New IO.MemoryStream(JPGFile)
        Return New Bitmap(iStream)
        iStream.Dispose()
      ElseIf Compression = 1 And (PictureType = 3 Or PictureType = 4) Then
        Dim ColorPallete(255) As UInt32
        ColorPallete = dFile.ReadUInt32Array(256)
        Dim MM_Width As Integer = Math.Floor(Width / (Item + 1))
        Dim MM_Height As Integer = Math.Floor(Height / (Item + 1))
        Dim bTmp(MiniMapSize(Item)) As Byte
        Array.Copy(bFile, MiniMapOffset(Item), bTmp, 0, MiniMapSize(Item))
        Dim dMiniMap As New DataReader(bTmp)
        Dim IndexList(MM_Width * MM_Height - 1) As Byte
        Dim AlphaList(MM_Width * MM_Height - 1) As Byte
        IndexList = dMiniMap.ReadByteArray(MM_Width * MM_Height)
        AlphaList = dMiniMap.ReadByteArray(MM_Width * MM_Height)
        Dim img As Image = New Bitmap(MM_Width, MM_Height)
        Dim g As Graphics = Graphics.FromImage(img)
        g.Clear(cBG)
        For J As Integer = 0 To MM_Width * MM_Height - 1
          Dim bColor As New SolidBrush(Color.FromArgb(AlphaList(J), Color.FromArgb(ColorPallete(IndexList(J)))))
          Dim rPos As Rectangle = New Rectangle(J Mod MM_Width, J \ MM_Width, 1, 1)
          g.FillRectangle(bColor, rPos)
        Next J
        g.Dispose()
        Return img
      ElseIf Compression = 1 And (PictureType = 5) Then
        Dim ColorPallete(255) As UInt32
        ColorPallete = dFile.ReadUInt32Array(256)
        Dim MM_Width As Integer = Math.Floor(Width / (Item + 1))
        Dim MM_Height As Integer = Math.Floor(Height / (Item + 1))
        Dim bTmp(MiniMapSize(Item)) As Byte
        Array.Copy(bFile, MiniMapOffset(Item), bTmp, 0, MiniMapSize(Item))
        Dim dMiniMap As New DataReader(bTmp)
        Dim IndexList(MM_Width * MM_Height - 1) As Byte
        IndexList = dMiniMap.ReadByteArray(MM_Width * MM_Height)
        Dim img As Image = New Bitmap(MM_Width, MM_Height)
        Dim g As Graphics = Graphics.FromImage(img)
        g.Clear(cBG)
        For J As Integer = 0 To MM_Width * MM_Height - 1
          Dim bColor As New SolidBrush(Color.FromArgb(&HFF000000 Xor ColorPallete(IndexList(J))))
          Dim rPos As Rectangle = New Rectangle(J Mod MM_Width, J \ MM_Width, 1, 1)
          g.FillRectangle(bColor, rPos)
        Next J
        g.Dispose()
        Return img
      Else
        Return New Bitmap(1, 1)
      End If
    Else
      Return New Bitmap(1, 1)
    End If
  End Function
End Module