﻿Public Class clsCSS
  Public Class StyleList
    Implements ICollection(Of Style)
    Public Structure Style
      Public Name As String
      Public Value As String

      Public Sub New(sName As String)
        Name = sName
        Value = String.Empty
      End Sub

      Public Sub New(sName As String, sValue As String)
        Name = sName
        Value = sValue
      End Sub
    End Structure
    Private Styles() As Style
    Public Sub New()
      Erase Styles
    End Sub
    Public Sub New(Entries() As Style)
      Styles = Entries
    End Sub
    Public Sub Add(item As Style) Implements System.Collections.Generic.ICollection(Of Style).Add
      If Styles IsNot Nothing Then
        For I As Integer = 0 To Styles.Length - 1
          If String.Compare(Styles(I).Name, item.Name, True) = 0 Then
            Styles(I).Value = item.Value
            Exit Sub
          End If
        Next
        ReDim Preserve Styles(Styles.Length)
        Styles(Styles.Length - 1) = item
      Else
        ReDim Styles(0)
        Styles(0) = item
      End If
    End Sub
    Public Sub Add(Name As String, Value As String)
      Add(New Style(Name, Value))
    End Sub
    Public Sub Clear() Implements System.Collections.Generic.ICollection(Of Style).Clear
      Erase Styles
    End Sub
    Public Function Contains(item As Style) As Boolean Implements System.Collections.Generic.ICollection(Of Style).Contains
      Return Styles.Contains(item)
    End Function
    Public Function Contains(item As String) As Boolean
      Return Styles.Contains(New Style(item))
    End Function

    Public Sub CopyTo(array() As Style, arrayIndex As Integer) Implements System.Collections.Generic.ICollection(Of Style).CopyTo
      System.Array.Copy(Styles, 0, array, arrayIndex, array.Length - arrayIndex)
    End Sub
    Public ReadOnly Property Count As Integer Implements System.Collections.Generic.ICollection(Of Style).Count
      Get
        If Styles Is Nothing Then Return 0
        Return Styles.Length
      End Get
    End Property
    Public ReadOnly Property IsReadOnly As Boolean Implements System.Collections.Generic.ICollection(Of Style).IsReadOnly
      Get
        Return False
      End Get
    End Property
    Public Function Remove(item As Style) As Boolean Implements System.Collections.Generic.ICollection(Of Style).Remove
      For I As Integer = 0 To Styles.Length - 1
        If String.Compare(Styles(I).Name, item.Name, True) = 0 Then
          For J As Integer = I + 1 To Styles.Length - 1
            Styles(J - 1) = Styles(J)
          Next
          ReDim Preserve Styles(Styles.Length - 2)
          Return True
        End If
      Next
      Return False
    End Function
    Public Function Remove(Name As String) As Boolean
      Return Remove(New Style(Name))
    End Function
    Public Function Find(Name As String) As Style
      If Styles Is Nothing Then Return New Style(Nothing, Nothing)
      Return Array.Find(Styles.ToArray, Function(Style As Style) String.Compare(Style.Name, Name, True) = 0)
    End Function
    Public Function GetEnumerator() As System.Collections.Generic.IEnumerator(Of Style) Implements System.Collections.Generic.IEnumerable(Of Style).GetEnumerator
      Return Styles.AsEnumerable.GetEnumerator
    End Function
    Public Function GetEnumerator1() As System.Collections.IEnumerator Implements System.Collections.IEnumerable.GetEnumerator
      Return Styles.AsEnumerable.GetEnumerator
    End Function
  End Class
  Public Class ElementList
    Implements ICollection(Of Element)
    Public Structure Element
      Public Name As String
      Public Styles As StyleList

      Public Sub New(sName As String)
        Name = sName
      End Sub

      Public Sub New(sName As String, sStyles As StyleList)
        Name = sName
        Styles = sStyles
      End Sub
    End Structure
    Private Items() As Element
    Public Sub Add(item As Element) Implements System.Collections.Generic.ICollection(Of Element).Add
      If Items IsNot Nothing Then
        For I As Integer = 0 To Items.Length - 1
          If String.Compare(Items(I).Name, item.Name, True) = 0 Then
            For Each style In item.Styles
              Items(I).Styles.Add(style)
            Next
            Exit Sub
          End If
        Next
        ReDim Preserve Items(Items.Length)
        Items(Items.Length - 1) = item
      Else
        ReDim Items(0)
        Items(0) = item
      End If
    End Sub
    Public Sub Add(Name As String, Styles As StyleList)
      Add(New Element(Name, Styles))
    End Sub
    Public Sub Clear() Implements System.Collections.Generic.ICollection(Of Element).Clear
      Erase Items
    End Sub
    Public Function Contains(item As Element) As Boolean Implements System.Collections.Generic.ICollection(Of Element).Contains
      Return Items.Contains(item)
    End Function
    Public Sub CopyTo(array() As Element, arrayIndex As Integer) Implements System.Collections.Generic.ICollection(Of Element).CopyTo
      System.Array.Copy(Items, 0, array, arrayIndex, array.Length - arrayIndex)
    End Sub
    Public ReadOnly Property Count As Integer Implements System.Collections.Generic.ICollection(Of Element).Count
      Get
        If Items Is Nothing Then Return 0
        Return Items.Length
      End Get
    End Property
    Public ReadOnly Property IsReadOnly As Boolean Implements System.Collections.Generic.ICollection(Of Element).IsReadOnly
      Get
        Return False
      End Get
    End Property
    Public Function Remove(item As Element) As Boolean Implements System.Collections.Generic.ICollection(Of Element).Remove
      For I As Integer = 0 To Items.Length - 1
        If String.Compare(Items(I).Name, item.Name, True) = 0 Then
          For J As Integer = I + 1 To Items.Length - 1
            Items(J - 1) = Items(J)
          Next
          ReDim Preserve Items(Items.Length - 2)
          Return True
        End If
      Next
      Return False
    End Function
    Public Function Find(Name As String) As Element
      If Items Is Nothing Then Return New Element(Nothing, New StyleList(Nothing))
      Return Array.Find(Items.ToArray, Function(Element As Element) String.Compare(Element.Name, Name, True) = 0)
    End Function
    Public Function GetEnumerator() As System.Collections.Generic.IEnumerator(Of Element) Implements System.Collections.Generic.IEnumerable(Of Element).GetEnumerator
      If Items Is Nothing Then Return Nothing
      Return Items.AsEnumerable.GetEnumerator
    End Function
    Public Function GetEnumerator1() As System.Collections.IEnumerator Implements System.Collections.IEnumerable.GetEnumerator
      Return Items.GetEnumerator
    End Function
  End Class
  Public Elements As ElementList
  Public Sub New()
    Elements = New ElementList
  End Sub
  Public Sub New(CSSText As String)
    If String.IsNullOrEmpty(CSSText) Then
      Elements = New ElementList
    Else
      Dim Chunks() As String = CSSText.Split("}"c)
      Elements = New ElementList
      For Each chunk In Chunks
        If chunk.Contains("{"c) Then
          Dim sTitle As String = chunk.Split("{"c)(0).Trim(vbCr, vbLf, vbTab, " "c)
          Dim sEntries() As String = chunk.Split("{"c)(1).Split(";"c)
          Dim lStyles As New StyleList
          For Each entry In sEntries
            If entry.Contains(":") Then
              lStyles.Add(New StyleList.Style(Split(entry, ":", 2)(0).Trim(vbCr, vbLf, vbTab, " "c), Split(entry, ":", 2)(1).Trim(vbCr, vbLf, vbTab, " "c).TrimEnd(";"c)))
            End If
          Next
          Elements.Add(New ElementList.Element(sTitle, lStyles))
        End If
      Next
    End If
  End Sub
  Public Overrides Function ToString() As String
    Dim sCSS As String = String.Empty
    If Elements.Count = 0 Then Return Nothing
    For Each el In Elements
      sCSS &= el.Name & vbNewLine & "{" & vbNewLine
      For Each Style In el.Styles
        sCSS &= " " & Style.Name & ": " & Style.Value & ";" & vbNewLine
      Next
      sCSS &= "}" & vbNewLine
    Next
    Return sCSS
  End Function
End Class