﻿Public Class MPQ
  Implements IDisposable
  Private Class NativeMethods
    Public Declare Function SFileCloseArchive Lib "StormLib.dll" (hMPQ As Integer) As Boolean
    Public Declare Function SFileCloseFile Lib "StormLib.dll" (hFile As Integer) As Boolean
    Public Declare Function SFileGetFileSize Lib "StormLib.dll" (hFile As Integer, ByRef lpFileSizeHigh As Integer) As Integer
    Public Declare Function SFileOpenArchive Lib "StormLib.dll" (lpFileName As String, dwPriority As Integer, dwFlags As Integer, ByRef hMPQ As Integer) As Boolean
    Public Declare Function SFileOpenFileEx Lib "StormLib.dll" (hMPQ As Integer, lpFileName As String, dwSearchScope As Integer, ByRef hFile As Integer) As Boolean
    Public Declare Function SFileReadFile Lib "StormLib.dll" (hFile As Integer, lpBuffer() As Byte, nNumberOfBYTEsToRead As Integer, ByRef lpNumberOfBYTEsRead As Integer, lpOverlapped As Integer) As Boolean
  End Class
  Private hMPQ As Integer
  Private cArchive As String
  Public Sub New()
    hMPQ = 0
  End Sub
  Public Sub New(strArchive As String)
    cArchive = strArchive
    OpenArchive()
  End Sub
  Public Property Archive As String
    Get
      Return cArchive
    End Get
    Set(value As String)
      cArchive = value
      If hMPQ <> 0 Then
        NativeMethods.SFileCloseArchive(hMPQ)
        hMPQ = 0
      End If
      If value <> String.Empty Then OpenArchive()
    End Set
  End Property
  Private Sub OpenArchive()
    NativeMethods.SFileOpenArchive(cArchive, 0, 0, hMPQ)
  End Sub
  Public Function GetListFile() As String
    Dim hFile As Integer
    Dim lngSize As Integer
    Dim dwBytes As Integer
    Dim bBuffer() As Byte
    If hMPQ = 0 Then Return String.Empty
    Try
      NativeMethods.SFileOpenFileEx(hMPQ, "(listfile)", 0, hFile)
      If hFile <> 0 Then
        lngSize = NativeMethods.SFileGetFileSize(hFile, 0)
        ReDim bBuffer(0 To lngSize - 1)
        NativeMethods.SFileReadFile(hFile, bBuffer, lngSize, dwBytes, &H0)
        If Not dwBytes = lngSize Then Err.Raise(vbObjectError + &H3, cArchive, "MPQ Corrupt (" & dwBytes & " / " & lngSize & ").")
        Return System.Text.Encoding.GetEncoding(LATIN_1).GetString(bBuffer)
      Else
        Return String.Empty
      End If
    Catch e As Exception
      Return String.Empty
    Finally
      If hFile <> 0 Then NativeMethods.SFileCloseFile(hFile)
    End Try
    Return String.Empty
  End Function
  Public Function OpenFile(sName As String) As Byte()
    Dim hFile As Integer
    Dim lngSize As Integer
    Dim dwBytes As Integer
    Dim bBuffer() As Byte
    If hMPQ = 0 Then Err.Raise(vbObjectError + &H1, cArchive, "No MPQ Open.")
    Try
      NativeMethods.SFileOpenFileEx(hMPQ, sName, 0, hFile)
      If hFile <> 0 Then
        lngSize = NativeMethods.SFileGetFileSize(hFile, 0)
        ReDim bBuffer(0 To lngSize - 1)
        NativeMethods.SFileReadFile(hFile, bBuffer, lngSize, dwBytes, &H0)
        If Not dwBytes = lngSize Then Err.Raise(vbObjectError + &H3, cArchive, "MPQ Corrupt (" & dwBytes & " / " & lngSize & ").")
        Return bBuffer
      Else
        Err.Raise(vbObjectError + &H5, cArchive, "File does not exist in MPQ.")
      End If
    Catch e As Exception
      Err.Raise(vbObjectError + &H2, cArchive, e.Message)
    Finally
      If hFile <> 0 Then NativeMethods.SFileCloseFile(hFile)
    End Try
    Return Nothing
  End Function
  Public Enum MPQSaveError
    Success
    NoMPQ
    MPQCorrupt
    FileNotFound
    UnknownFailure
  End Enum
  Public Function SaveToPath(sName As String, sDir As String) As MPQSaveError
    Dim hFile As Integer
    Dim lngSize As Integer
    Dim dwBytes As Integer
    Dim bBuffer() As Byte
    If hMPQ = 0 Then Return MPQSaveError.NoMPQ
    Try
      NativeMethods.SFileOpenFileEx(hMPQ, sName, 0, hFile)
      If hFile <> 0 Then
        lngSize = NativeMethods.SFileGetFileSize(hFile, 0)
        ReDim bBuffer(0 To lngSize - 1)
        NativeMethods.SFileReadFile(hFile, bBuffer, lngSize, dwBytes, &H0)
        If Not dwBytes = lngSize Then Return MPQSaveError.MPQCorrupt
        If Not My.Computer.FileSystem.DirectoryExists(sDir) Then My.Computer.FileSystem.CreateDirectory(sDir)
        My.Computer.FileSystem.WriteAllBytes(sDir & "\" & IO.Path.GetFileName(sName), bBuffer, False)
        Return MPQSaveError.Success
      Else
        Return MPQSaveError.FileNotFound
      End If
    Catch e As Exception
      Return MPQSaveError.UnknownFailure
    Finally
      If hFile <> 0 Then NativeMethods.SFileCloseFile(hFile)
    End Try
  End Function
#Region "IDisposable Support"
  Private disposedValue As Boolean
  Protected Overridable Sub Dispose(disposing As Boolean)
    If Not Me.disposedValue Then
      If disposing Then
        cArchive = Nothing
      End If
      If hMPQ <> 0 Then NativeMethods.SFileCloseArchive(hMPQ)
      hMPQ = 0
    End If
    Me.disposedValue = True
  End Sub
  Public Sub Dispose() Implements IDisposable.Dispose
    Dispose(True)
    GC.SuppressFinalize(Me)
  End Sub
#End Region
End Class