﻿Public Class ChessBoard
  Private colorSelected As Color = Color.SandyBrown
  Private colorMoves As Color = Color.PeachPuff
  Private colorCheck As Color = Color.Tomato
  Private colorBlack As Color = Color.ForestGreen
  Private colorWhite As Color = Color.Ivory
  Private cChess As clsChess
  Private bMyTurn As Boolean
  Private bMyColor As Color
  Private bMySize As Byte
  Private ResMan As New Resources.ResourceManager("Entelechy.frmChessBoard", Reflection.[Assembly].GetExecutingAssembly)
  Class MoveArgs
    Inherits EventArgs
    Public sMove As String
    Public Sub New(Move As String)
      sMove = Move
    End Sub
  End Class
  Event MadeMove(sender As Object, e As MoveArgs)
  Event NewGame(sender As Object, e As EventArgs)
  Event Resign(sender As Object, e As EventArgs)
  Event Draw(sender As Object, e As EventArgs)
  Event SaveAndQuit(sender As Object, e As EventArgs)
  Public sOpponent As String
  Public ReadOnly Property IsPlaying As Boolean
    Get
      Return cChess IsNot Nothing
    End Get
  End Property
  Public ReadOnly Property MyTurn As Boolean
    Get
      Return bMyTurn
    End Get
  End Property
  Public ReadOnly Property MyColor As Color
    Get
      Return bMyColor
    End Get
  End Property
  Public Property SquareSize As Byte
    Get
      Return bMySize
    End Get
    Set(value As Byte)
      bMySize = value
      DrawBoard()
    End Set
  End Property
  Public Sub New()
    InitializeComponent()
    bMySize = 32
    For Each pCell In pnlBoard.Controls
      If pCell.GetType.Equals(GetType(PictureBox)) Then
        If pCell.BackColor = Color.Black Then pCell.Tag = 1
        If pCell.BackColor = Color.White Then pCell.Tag = 0
        pCell.Size = New Size(bMySize, bMySize)
      End If
    Next
    lblMove.Text = String.Empty
    lblTurn.Text = String.Empty
  End Sub
  Public Sub setColors(BlackTiles As Color, WhiteTiles As Color, Selected As Color, Moves As Color, Check As Color)
    colorSelected = Selected
    colorMoves = Moves
    colorCheck = Check
    colorBlack = BlackTiles
    colorWhite = WhiteTiles
    If cChess IsNot Nothing Then
      cChess.cWhiteTiles = WhiteTiles
      cChess.cBlackTiles = BlackTiles
    End If
  End Sub
  Private Function GetPieceIco(pColor As Color, pType As clsChess.PieceType) As Icon
    Dim PieceDLL As Reflection.Assembly = Reflection.Assembly.LoadFile(Application.StartupPath & "\ChessPieces.dll")
    Dim sStream As String = "ChessPieces.I" & bMySize & "_" & IIf(pColor = Color.White, "W", "B") & "_" & pType.ToString & ".ico"
    Dim pStream As IO.Stream = PieceDLL.GetManifestResourceStream(sStream)
    If pStream IsNot Nothing Then Return New Icon(pStream)
    Return My.Resources.disconnect
  End Function
  Private Function GetPieceCur(pColor As Color, pType As clsChess.PieceType) As Cursor
    Dim PieceDLL As Reflection.Assembly = Reflection.Assembly.LoadFile(Application.StartupPath & "\ChessPieces.dll")
    Dim pStream As IO.Stream = PieceDLL.GetManifestResourceStream("ChessPieces.I" & bMySize & "_" & IIf(pColor = Color.White, "W", "B") & "_" & pType.ToString & ".ico")
    Return New Cursor(pStream)
  End Function
  Private Sub DrawBoard()
    pnlBoard.SuspendLayout()
    If cChess IsNot Nothing Then
      For R As Integer = 1 To 8
        For C As Integer = 1 To 8
          Dim cPos As New clsChess.PiecePos(R, C)
          Dim pCell As PictureBox = pnlBoard.Controls.Find("pct" & cPos.ToString, False)(0)
          pCell.Size = New Size(bMySize, bMySize)
          'pCell.Width = 48
          'pCell.Height = 48
          Dim tPiece As clsChess.PieceType = cChess.GetPieceType(cPos)
          Dim cPiece As Color = cChess.GetPieceColor(cPos)
          If tPiece = Entelechy.clsChess.PieceType.Nothing Then
            pCell.Image = Nothing
            pCell.SizeMode = PictureBoxSizeMode.Normal
            pCell.BackColor = IIf(pCell.Tag = 1, colorBlack, colorWhite)
            ttPiece.SetToolTip(pCell, String.Empty)
          Else
            pCell.Image = GetPieceIco(cPiece, tPiece).ToBitmap
            pCell.SizeMode = PictureBoxSizeMode.Normal
            ttPiece.SetToolTip(pCell, tPiece.ToString)
            If tPiece = Entelechy.clsChess.PieceType.King Then
              If cChess.isInCheck(cPiece = Color.White) Then
                pCell.BackColor = colorCheck
              Else
                pCell.BackColor = IIf(pCell.Tag = 1, colorBlack, colorWhite)
              End If
            Else
              pCell.BackColor = IIf(pCell.Tag = 1, colorBlack, colorWhite)
            End If
          End If
          pCell.SizeMode = PictureBoxSizeMode.CenterImage
        Next
      Next
    Else
      For Each pCell In pnlBoard.Controls
        If pCell.GetType = GetType(PictureBox) Then
          pCell.Size = New Size(bMySize, bMySize)
          pCell.SizeMode = PictureBoxSizeMode.StretchImage
        End If
      Next
    End If
    pnlBoard.ResumeLayout(True)
  End Sub
  Public Sub StartGame(asWhite As Boolean)
    If Not asWhite Then
      For Each pCell In pnlBoard.Controls
        If pCell.GetType.Equals(GetType(PictureBox)) Then
          Dim pBox As PictureBox = pCell
          Dim pName As String = "pct" & New clsChess.PiecePos(pnlBoard.GetRow(pBox) + 1, pnlBoard.GetColumn(pBox) + 1).ToString
          pBox.Name = pName
        End If
      Next
    Else
      For Each pCell In pnlBoard.Controls
        If pCell.GetType.Equals(GetType(PictureBox)) Then
          Dim pBox As PictureBox = pCell
          Dim pName As String = "pct" & New clsChess.PiecePos(8 - (pnlBoard.GetRow(pBox)), pnlBoard.GetColumn(pBox) + 1).ToString
          pBox.Name = pName
        End If
      Next
    End If
    cChess = New clsChess
    cChess.cBlackTiles = colorBlack
    cChess.cWhiteTiles = colorWhite
    DrawBoard()
    cmdDraw.Enabled = True
    cmdResign.Enabled = True
    cmdSave.Enabled = True
    bMyColor = IIf(asWhite, Color.White, Color.Black)
    bMyTurn = asWhite
    If bMyTurn Then
      lblTurn.Text = ResMan.GetString("lblTurn.Yours")
    Else
      If bMyColor = Color.White Then
        lblTurn.Text = ResMan.GetString("lblTurn.Black")
      Else
        lblTurn.Text = ResMan.GetString("lblTurn.White")
      End If
    End If
    lblMove.Text = String.Empty
  End Sub
  Public Sub ResumeGame()
    cChess = New clsChess
    Dim asWhite As Boolean
    If Not cChess.LoadGame(AppStore & "\Chess\" & ValidFile(sOpponent) & ".chs", asWhite, bMyTurn) Then
      Using cMsg As New MessageFrame
        cMsg.MsgBox(Me, ResString(ResMan.GetString("MSG_CORRUPTSAVE"), sOpponent), MsgBoxStyle.Critical, ResMan.GetString("MSG_CORRUPTSAVE_TITLE"))
      End Using
      My.Computer.FileSystem.DeleteFile(AppStore & "\Chess\" & ValidFile(sOpponent) & ".chs")
      Exit Sub
    End If
    If Not asWhite Then
      For Each pCell In pnlBoard.Controls
        If pCell.GetType.Equals(GetType(PictureBox)) Then
          Dim pBox As PictureBox = pCell
          Dim pName As String = "pct" & New clsChess.PiecePos(pnlBoard.GetRow(pBox) + 1, pnlBoard.GetColumn(pBox) + 1).ToString
          pBox.Name = pName
        End If
      Next
    Else
      For Each pCell In pnlBoard.Controls
        If pCell.GetType.Equals(GetType(PictureBox)) Then
          Dim pBox As PictureBox = pCell
          Dim pName As String = "pct" & New clsChess.PiecePos(8 - (pnlBoard.GetRow(pBox)), pnlBoard.GetColumn(pBox) + 1).ToString
          pBox.Name = pName
        End If
      Next
    End If
    cChess.cBlackTiles = colorBlack
    cChess.cWhiteTiles = colorWhite
    DrawBoard()
    cmdDraw.Enabled = True
    cmdResign.Enabled = True
    cmdSave.Enabled = True
    bMyColor = IIf(asWhite, Color.White, Color.Black)
    If bMyTurn Then
      lblTurn.Text = ResMan.GetString("lblTurn.Yours")
    Else
      If bMyColor = Color.White Then
        lblTurn.Text = ResMan.GetString("lblTurn.Black")
      Else
        lblTurn.Text = ResMan.GetString("lblTurn.White")
      End If
    End If
    lblMove.Text = String.Empty
    My.Computer.FileSystem.DeleteFile(AppStore & "\Chess\" & ValidFile(sOpponent) & ".chs")
  End Sub
  Public Sub FromString(BoardData As String)
    cChess = New clsChess
    Dim asWhite As Boolean
    If BoardData.Contains("You are White.") Then
      asWhite = True
    Else
      asWhite = False
    End If
    bMyTurn = BoardData.Contains("It is your turn.")
    BoardData = BoardData.Substring(0, BoardData.Length - 16)
    If String.IsNullOrEmpty(BoardData) Then Exit Sub
    cChess.StringToBoard(BoardData)
    If Not asWhite Then
      For Each pCell In pnlBoard.Controls
        If pCell.GetType.Equals(GetType(PictureBox)) Then
          Dim pBox As PictureBox = pCell
          Dim pName As String = "pct" & New clsChess.PiecePos(pnlBoard.GetRow(pBox) + 1, pnlBoard.GetColumn(pBox) + 1).ToString
          pBox.Name = pName
        End If
      Next
    Else
      For Each pCell In pnlBoard.Controls
        If pCell.GetType.Equals(GetType(PictureBox)) Then
          Dim pBox As PictureBox = pCell
          Dim pName As String = "pct" & New clsChess.PiecePos(8 - (pnlBoard.GetRow(pBox)), pnlBoard.GetColumn(pBox) + 1).ToString
          pBox.Name = pName
        End If
      Next
    End If
    cChess.cBlackTiles = colorBlack
    cChess.cWhiteTiles = colorWhite
    DrawBoard()
    cmdDraw.Enabled = True
    cmdResign.Enabled = True
    cmdSave.Enabled = True
    bMyColor = IIf(asWhite, Color.White, Color.Black)
    If bMyTurn Then
      lblTurn.Text = ResMan.GetString("lblTurn.Yours")
    Else
      If bMyColor = Color.White Then
        lblTurn.Text = ResMan.GetString("lblTurn.Black")
      Else
        lblTurn.Text = ResMan.GetString("lblTurn.White")
      End If
    End If
    lblMove.Text = String.Empty
  End Sub
  Public Function OpponentMove(Move As String) As Boolean
    Dim pStart, pEnd As clsChess.PiecePos
    Dim pPromote As clsChess.PieceType = clsChess.PieceType.Nothing
    If Move.StartsWith("0-0-0") Or Move.ToUpper.StartsWith("O-O-O") Or Move.ToUpper.StartsWith("0-0L") Or Move.ToUpper.StartsWith("O-OL") Then
      'King Castle Queen Side
      If bMyColor = Color.White Then
        pStart = New clsChess.PiecePos("E8")
        pEnd = New clsChess.PiecePos("C8")
      Else
        pStart = New clsChess.PiecePos("E1")
        pEnd = New clsChess.PiecePos("C1")
      End If
    ElseIf Move.StartsWith("0-0") Or Move.ToUpper.StartsWith("O-O") Or Move.ToUpper.StartsWith("0-0R") Or Move.ToUpper.StartsWith("O-O-R") Then
      'King Castle King Side
      If bMyColor = Color.White Then
        pStart = New clsChess.PiecePos("E8")
        pEnd = New clsChess.PiecePos("G8")
      Else
        pStart = New clsChess.PiecePos("E1")
        pEnd = New clsChess.PiecePos("G1")
      End If
    Else
      If Move.Length > 1 Then
        If IsNumeric(Move(1)) Then
          'pawn
          pStart = New clsChess.PiecePos(Move.Substring(0, 2))
          pEnd = New clsChess.PiecePos(Move.Substring(3, 2))
          Do While Move.EndsWith(":") Or Move.ToUpper.EndsWith("X") Or Move.EndsWith("+") Or Move.EndsWith("#") Or Move.EndsWith("?") Or Move.EndsWith("!")
            Move = Move.Substring(0, Move.Length - 1)
          Loop
          If Move.ToUpper.EndsWith("Q") Then
            pPromote = clsChess.PieceType.Queen
          ElseIf Move.ToUpper.EndsWith("B") Then
            pPromote = clsChess.PieceType.Bishop
          ElseIf Move.ToUpper.EndsWith("N") Then
            pPromote = clsChess.PieceType.Knight
          ElseIf Move.ToUpper.EndsWith("R") Then
            pPromote = clsChess.PieceType.Rook
          End If
        ElseIf IsNumeric(Move(2)) Then
          'Else
          pStart = New clsChess.PiecePos(Move.Substring(1, 2))
          pEnd = New clsChess.PiecePos(Move.Substring(4, 2))
        End If
      End If
    End If
    'Debug.Print(Move & " [" & pStart.ToString & " " & pEnd.ToString & " (" & pPromote.ToString & ")]")
    Dim moveRet = cChess.MovePiece(pStart, pEnd, False, bMyColor, , pPromote, bMySize)
    DrawBoard()
    lblTurn.Text = ResMan.GetString("lblTurn.Yours")
    lblMove.Text = Move
    If CBool(moveRet And clsChess.MoveTypes.Move) Then
      HandleMove(moveRet, False)
      bMyTurn = True
      Return True
    End If
    Return False
  End Function
  Private bMoving As Boolean
  Private selPiece As clsChess.PiecePos
  Private Sub pctSquare_Click(sender As System.Object, e As System.EventArgs) Handles pctA1.Click, pctA2.Click, pctA3.Click, pctA4.Click, pctA5.Click, pctA6.Click, pctA7.Click, pctA8.Click,
                                                                                      pctB1.Click, pctB2.Click, pctB3.Click, pctB4.Click, pctB5.Click, pctB6.Click, pctB7.Click, pctB8.Click,
                                                                                      pctC1.Click, pctC2.Click, pctC3.Click, pctC4.Click, pctC5.Click, pctC6.Click, pctC7.Click, pctC8.Click,
                                                                                      pctD1.Click, pctD2.Click, pctD3.Click, pctD4.Click, pctD5.Click, pctD6.Click, pctD7.Click, pctD8.Click,
                                                                                      pctE1.Click, pctE2.Click, pctE3.Click, pctE4.Click, pctE5.Click, pctE6.Click, pctE7.Click, pctE8.Click,
                                                                                      pctF1.Click, pctF2.Click, pctF3.Click, pctF4.Click, pctF5.Click, pctF6.Click, pctF7.Click, pctF8.Click,
                                                                                      pctG1.Click, pctG2.Click, pctG3.Click, pctG4.Click, pctG5.Click, pctG6.Click, pctG7.Click, pctG8.Click,
                                                                                      pctH1.Click, pctH2.Click, pctH3.Click, pctH4.Click, pctH5.Click, pctH6.Click, pctH7.Click, pctH8.Click
    If cChess Is Nothing Then Exit Sub
    If Not bMyTurn Then Exit Sub
    Dim pCell As PictureBox = sender
    Dim cPos As New clsChess.PiecePos(pCell.Name.Substring(3))
    If bMoving Then
      Me.Cursor = Cursors.Default
      Dim cLastPos As clsChess.PiecePos = selPiece
      Dim pLastType As clsChess.PieceType = cChess.GetPieceType(selPiece)
      Dim moveRet = cChess.MovePiece(selPiece, cPos, True, bMyColor, , , bMySize)
      If CBool(moveRet And clsChess.MoveTypes.Move) Then
        Dim sMove As String = ""
        If CBool(moveRet And clsChess.MoveTypes.Castle) Then
          If cPos.sCol = "C"c Then
            sMove = "0-0-0"
          Else
            sMove = "0-0"
          End If
        Else
          Select Case pLastType
            Case clsChess.PieceType.King : sMove &= "K"c
            Case clsChess.PieceType.Queen : sMove &= "Q"c
            Case clsChess.PieceType.Rook : sMove &= "R"c
            Case clsChess.PieceType.Bishop : sMove &= "B"c
            Case clsChess.PieceType.Knight : sMove &= "N"c
            Case clsChess.PieceType.Pawn : sMove &= ""
          End Select
          sMove &= cLastPos.ToString.ToLower
          If CBool(moveRet And clsChess.MoveTypes.Capture) Then
            sMove &= "x"
          Else
            sMove &= " "
          End If
          sMove &= cPos.ToString.ToLower
          If CBool(moveRet And clsChess.MoveTypes.Promote) Then
            Select Case cChess.GetPieceType(cPos)
              Case clsChess.PieceType.Queen : sMove &= "Q"
              Case clsChess.PieceType.Rook : sMove &= "R"
              Case clsChess.PieceType.Bishop : sMove &= "B"
              Case clsChess.PieceType.Knight : sMove &= "N"
            End Select
          End If
          If CBool(moveRet And clsChess.MoveTypes.Checkmate) Then
            sMove &= "#"
          ElseIf CBool(moveRet And clsChess.MoveTypes.Check) Then
            sMove &= "+"
          End If
        End If
        lblMove.Text = sMove
        RaiseEvent MadeMove(Me, New MoveArgs(sMove))
        HandleMove(moveRet, True)
        bMyTurn = False
        If bMyColor = Color.White Then
          lblTurn.Text = ResMan.GetString("lblTurn.Black")
        Else
          lblTurn.Text = ResMan.GetString("lblTurn.White")
        End If
      Else
        lblMove.Text = lblMove.Tag
      End If
      bMoving = False
      selPiece = Nothing
      DrawBoard()
    Else
      Dim tPiece As clsChess.PieceType = cChess.GetPieceType(cPos)
      If Not tPiece = clsChess.PieceType.Nothing And cChess.GetPieceColor(cPos) = bMyColor Then
        lblMove.Tag = lblMove.Text
        Select Case tPiece
          Case clsChess.PieceType.King : lblMove.Text = "K" & cPos.ToString.ToLower
          Case clsChess.PieceType.Queen : lblMove.Text = "Q" & cPos.ToString.ToLower
          Case clsChess.PieceType.Rook : lblMove.Text = "R" & cPos.ToString.ToLower
          Case clsChess.PieceType.Bishop : lblMove.Text = "B" & cPos.ToString.ToLower
          Case clsChess.PieceType.Knight : lblMove.Text = "N" & cPos.ToString.ToLower
          Case clsChess.PieceType.Pawn : lblMove.Text = cPos.ToString.ToLower
        End Select
        selPiece = cPos
        bMoving = True
        pCell.BackColor = colorSelected
        Dim cMoves As clsChess.PiecePos() = cChess.GetPieceMoves(selPiece)
        For Each cMove As clsChess.PiecePos In cMoves
          Dim pMoveCell As PictureBox = pnlBoard.Controls.Find("pct" & cMove.ToString, False)(0)
          pMoveCell.BackColor = colorMoves
        Next
        Dim cPiece As Color = cChess.GetPieceColor(cPos)
        pCell.Image = Nothing
        Me.Cursor = GetPieceCur(cPiece, tPiece)
      End If
    End If
  End Sub
  Public Sub HandleMove(moveRet As clsChess.MoveTypes, NewPrompt As Boolean)
    If CBool(moveRet And clsChess.MoveTypes.Checkmate) Or CBool(moveRet And Entelechy.clsChess.MoveTypes.Draw) Then
      DrawBoard()
      cmdDraw.Enabled = False
      cmdResign.Enabled = False
      cmdSave.Enabled = False
      If NewPrompt Then
        Using mFrame As New MessageFrame
          If mFrame.MsgBox(Me, ResString(ResMan.GetString("MSG_NEWGAME"), Nothing), MsgBoxStyle.Question Or MsgBoxStyle.YesNo) = MsgBoxResult.Yes Then
            cChess = Nothing
            RaiseEvent NewGame(Me, New EventArgs)
          Else
            cChess = Nothing
          End If
        End Using
      Else
        Using mFrame As New MessageFrame
          If CBool(moveRet And Entelechy.clsChess.MoveTypes.Draw) Then
            mFrame.MsgBox(Me, ResMan.GetString("MSG_DRAW"), MsgBoxStyle.Information)
          Else
            mFrame.MsgBox(Me, ResMan.GetString("CHECKMATE"), MsgBoxStyle.Critical)
          End If
        End Using
        cChess = Nothing
      End If
    End If
  End Sub
  Private Sub cmdResign_Click(sender As System.Object, e As System.EventArgs) Handles cmdResign.Click
    If cChess IsNot Nothing Then
      Using mFrame As New MessageFrame
        If mFrame.MsgBox(Me, ResString(ResMan.GetString("MSG_RESIGN"), sOpponent), MsgBoxStyle.YesNo Or MsgBoxStyle.Question, ResMan.GetString("MSG_RESIGN_TITLE")) = MsgBoxResult.Yes Then RaiseEvent Resign(Me, New EventArgs)
      End Using
    Else
      Using mFrame As New MessageFrame : mFrame.MsgBox(Me, ResString(ResMan.GetString("MSG_NO_RESIGN")), MsgBoxStyle.Critical) : End Using
    End If
  End Sub
  Private Sub cmdDraw_Click(sender As System.Object, e As System.EventArgs) Handles cmdDraw.Click
    If cChess IsNot Nothing Then
      Using mFrame As New MessageFrame
        If mFrame.MsgBox(Me, ResString(ResMan.GetString("MSG_DRAW_REQUEST"), sOpponent), MsgBoxStyle.Question Or MsgBoxStyle.YesNo, ResMan.GetString("MSG_DRAW_REQUEST_TITLE")) = MsgBoxResult.Yes Then RaiseEvent Draw(Me, New EventArgs)
      End Using
    Else
      Using mFrame As New MessageFrame : mFrame.MsgBox(Me, ResString(ResMan.GetString("MSG_NO_DRAW")), MsgBoxStyle.Critical) : End Using
    End If
  End Sub
  Private Sub cmdSave_Click(sender As System.Object, e As System.EventArgs) Handles cmdSave.Click
    If cChess IsNot Nothing Then
      Using mFrame As New MessageFrame
        If mFrame.MsgBox(Me, ResString(ResMan.GetString("MSG_SAVE"), sOpponent), MsgBoxStyle.Question Or MsgBoxStyle.YesNo, ResMan.GetString("MSG_SAVE_TITLE")) = MsgBoxResult.Yes Then
          If Not My.Computer.FileSystem.DirectoryExists(AppStore & "\Chess\") Then My.Computer.FileSystem.CreateDirectory(AppStore & "\Chess\")
          cChess.SaveGame(AppStore & "\Chess\" & ValidFile(sOpponent) & ".chs", bMyColor = Color.White, bMyTurn)
          RaiseEvent SaveAndQuit(Me, New EventArgs)
        End If
      End Using
    Else
      Using mFrame As New MessageFrame : mFrame.MsgBox(Me, ResString(ResMan.GetString("MSG_NO_SAVE")), MsgBoxStyle.Critical) : End Using
    End If
  End Sub
  Public Overrides Function ToString() As String
    If cChess IsNot Nothing Then
      Return cChess.BoardToString
    Else
      Return "No Game"
    End If
  End Function
End Class